"""Fast algorithms for the densest subgraph problem"""

import math

import networkx as nx

__all__ = ["densest_subgraph"]


def _greedy_plus_plus(G, iterations):
    if G.number_of_edges() == 0:
        return 0.0, set()
    if iterations < 1:
        raise ValueError(
            f"The number of iterations must be an integer >= 1. Provided: {iterations}"
        )

    loads = {node: 0 for node in G.nodes}  # Load vector for Greedy++.
    best_density = 0.0  # Highest density encountered.
    best_subgraph = set()  # Nodes of the best subgraph found.

    for _ in range(iterations):
        # Initialize heap for fast access to minimum weighted degree.
        heap = nx.utils.BinaryHeap()

        # Compute initial weighted degrees and add nodes to the heap.
        for node, degree in G.degree:
            heap.insert(node, loads[node] + degree)
        # Set up tracking for current graph state.
        remaining_nodes = set(G.nodes)
        num_edges = G.number_of_edges()
        current_degrees = dict(G.degree)

        while remaining_nodes:
            num_nodes = len(remaining_nodes)

            # Current density of the (implicit) graph
            current_density = num_edges / num_nodes

            # Update the best density.
            if current_density > best_density:
                best_density = current_density
                best_subgraph = set(remaining_nodes)

            # Pop the node with the smallest weighted degree.
            node, _ = heap.pop()
            if node not in remaining_nodes:
                continue  # Skip nodes already removed.

            # Update the load of the popped node.
            loads[node] += current_degrees[node]

            # Update neighbors' degrees and the heap.
            for neighbor in G.neighbors(node):
                if neighbor in remaining_nodes:
                    current_degrees[neighbor] -= 1
                    num_edges -= 1
                    heap.insert(neighbor, loads[neighbor] + current_degrees[neighbor])

            # Remove the node from the remaining nodes.
            remaining_nodes.remove(node)

    return best_density, best_subgraph


def _fractional_peeling(G, b, x, node_to_idx, edge_to_idx):
    """
    Optimized fractional peeling using NumPy arrays.

    Parameters
    ----------
    G : networkx.Graph
        The input graph.
    b : numpy.ndarray
        Induced load vector.
    x : numpy.ndarray
        Fractional edge values.
    node_to_idx : dict
        Mapping from node to index.
    edge_to_idx : dict
        Mapping from edge to index.

    Returns
    -------
    best_density : float
        The best density found.
    best_subgraph : set
        The subset of nodes defining the densest subgraph.
    """
    heap = nx.utils.BinaryHeap()

    remaining_nodes = set(G.nodes)

    # Initialize heap with b values
    for idx, node in enumerate(G):
        heap.insert(node, b[idx])

    num_edges = G.number_of_edges()

    best_density = 0.0
    best_subgraph = set()

    while remaining_nodes:
        num_nodes = len(remaining_nodes)
        current_density = num_edges / num_nodes

        if current_density > best_density:
            best_density = current_density
            best_subgraph = set(remaining_nodes)

        # Pop the node with the smallest b
        node, _ = heap.pop()
        while node not in remaining_nodes:
            node, _ = heap.pop()  # Clean the heap from stale values

        # Update neighbors b values by subtracting fractional x value
        for neighbor in G.neighbors(node):
            if neighbor in remaining_nodes:
                neighbor_idx = node_to_idx[neighbor]
                # Take off fractional value
                b[neighbor_idx] -= x[edge_to_idx[(neighbor, node)]]
                num_edges -= 1
                heap.insert(neighbor, b[neighbor_idx])

        remaining_nodes.remove(node)  # peel off node

    return best_density, best_subgraph


def _fista(G, iterations):
    if G.number_of_edges() == 0:
        return 0.0, set()
    if iterations < 1:
        raise ValueError(
            f"The number of iterations must be an integer >= 1. Provided: {iterations}"
        )
    import numpy as np

    # 1. Node Mapping: Assign a unique index to each node and edge
    node_to_idx = {node: idx for idx, node in enumerate(G)}
    num_nodes = G.number_of_nodes()
    num_undirected_edges = G.number_of_edges()

    # 2. Edge Mapping: Assign a unique index to each bidirectional edge
    bidirectional_edges = [(u, v) for u, v in G.edges] + [(v, u) for u, v in G.edges]
    edge_to_idx = {edge: idx for idx, edge in enumerate(bidirectional_edges)}

    num_edges = len(bidirectional_edges)

    # 3. Reverse Edge Mapping: Map each (bidirectional) edge to its reverse edge index
    reverse_edge_idx = np.empty(num_edges, dtype=np.int32)
    for idx in range(num_undirected_edges):
        reverse_edge_idx[idx] = num_undirected_edges + idx
    for idx in range(num_undirected_edges, 2 * num_undirected_edges):
        reverse_edge_idx[idx] = idx - num_undirected_edges

    # 4. Initialize Variables as NumPy Arrays
    x = np.full(num_edges, 0.5, dtype=np.float32)
    y = x.copy()
    z = np.zeros(num_edges, dtype=np.float32)
    b = np.zeros(num_nodes, dtype=np.float32)  # Induced load vector
    tk = 1.0  # Momentum term

    # 5. Precompute Edge Source Indices
    edge_src_indices = np.array(
        [node_to_idx[u] for u, _ in bidirectional_edges], dtype=np.int32
    )

    # 6. Compute Learning Rate
    max_degree = max(deg for _, deg in G.degree)
    # 0.9 for floating point errs when max_degree is very large
    learning_rate = 0.9 / max_degree

    # 7. Iterative Updates
    for _ in range(iterations):
        # 7a. Update b: sum y over outgoing edges for each node
        b[:] = 0.0  # Reset b to zero
        np.add.at(b, edge_src_indices, y)  # b_u = \sum_{v : (u,v) \in E(G)} y_{uv}

        # 7b. Compute z, z_{uv} = y_{uv} - 2 * learning_rate * b_u
        z = y - 2.0 * learning_rate * b[edge_src_indices]

        # 7c. Update Momentum Term
        tknew = (1.0 + math.sqrt(1 + 4.0 * tk**2)) / 2.0

        # 7d. Update x in a vectorized manner, x_{uv} = (z_{uv} - z_{vu} + 1.0) / 2.0
        new_xuv = (z - z[reverse_edge_idx] + 1.0) / 2.0
        clamped_x = np.clip(new_xuv, 0.0, 1.0)  # Clamp x_{uv} between 0 and 1

        # Update y using the FISTA update formula (similar to gradient descent)
        y = (
            clamped_x
            + ((tk - 1.0) / tknew) * (clamped_x - x)
            + (tk / tknew) * (clamped_x - y)
        )

        # Update x
        x = clamped_x

        # Update tk, the momemntum term
        tk = tknew

    # Rebalance the b values! Otherwise performance is a bit suboptimal.
    b[:] = 0.0
    np.add.at(b, edge_src_indices, x)  # b_u = \sum_{v : (u,v) \in E(G)} x_{uv}

    # Extract the actual (approximate) dense subgraph.
    return _fractional_peeling(G, b, x, node_to_idx, edge_to_idx)


ALGORITHMS = {"greedy++": _greedy_plus_plus, "fista": _fista}


@nx.utils.not_implemented_for("directed")
@nx.utils.not_implemented_for("multigraph")
@nx._dispatchable
def densest_subgraph(G, iterations=1, *, method="fista"):
    r"""Returns an approximate densest subgraph for a graph `G`.

    This function runs an iterative algorithm to find the densest subgraph,
    and returns both the density and the subgraph. For a discussion on the
    notion of density used and the different algorithms available on
    networkx, please see the Notes section below.

    Parameters
    ----------
    G : NetworkX graph
        Undirected graph.

    iterations : int, optional (default=1)
        Number of iterations to use for the iterative algorithm. Can be
        specified positionally or as a keyword argument.

    method : string, optional (default='fista')
        The algorithm to use to approximate the densest subgraph. Supported
        options: 'greedy++' by Boob et al. [2]_ and 'fista' by Harb et al. [3]_.
        Must be specified as a keyword argument. Other inputs produce a
        ValueError.

    Returns
    -------
    d : float
        The density of the approximate subgraph found.

    S : set
        The subset of nodes defining the approximate densest subgraph.

    Examples
    --------
    >>> G = nx.star_graph(4)
    >>> nx.approximation.densest_subgraph(G, iterations=1)
    (0.8, {0, 1, 2, 3, 4})

    Notes
    -----
    **Problem Definition:**
    The densest subgraph problem (DSG) asks to find the subgraph
    $S \subseteq V(G)$ with maximum density. For a subset of the nodes of
    $G$, $S \subseteq V(G)$, define $E(S) = \{ (u,v) : (u,v)\in E(G),
    u\in S, v\in S \}$ as the set of edges with both endpoints in $S$.
    The density of $S$ is defined as $|E(S)|/|S|$, the ratio between the
    edges in the subgraph $G[S]$ and the number of nodes in that subgraph.
    Note that this is different from the standard graph theoretic definition
    of density, defined as $\frac{2|E(S)|}{|S|(|S|-1)}$, for historical
    reasons.

    **Exact Algorithms:**
    The densest subgraph problem is polynomial time solvable using maximum
    flow, commonly referred to as Goldberg's algorithm. However, the
    algorithm is quite involved. It first binary searches on the optimal
    density, $d^\ast$. For a guess of the density $d$, it sets up a flow
    network $G'$ with size $O(m)$. The maximum flow solution either
    informs the algorithm that no subgraph with density $d$ exists, or it
    provides a subgraph with density at least $d$. However, this is
    inherently bottlenecked by the maximum flow algorithm. For example, [2]_
    notes that Goldberg’s algorithm was not feasible on many large graphs
    even though they used a highly optimized maximum flow library.

    **Charikar's Greedy Peeling:**
    While exact solution algorithms are quite involved, there are several
    known approximation algorithms for the densest subgraph problem.

    Charikar [1]_ described a very simple 1/2-approximation algorithm for DSG
    known as the greedy "peeling" algorithm. The algorithm creates an
    ordering of the nodes as follows. The first node $v_1$ is the one with
    the smallest degree in $G$ (ties broken arbitrarily). It selects
    $v_2$ to be the smallest degree node in $G \setminus v_1$. Letting
    $G_i$ be the graph after removing $v_1, ..., v_i$ (with $G_0=G$),
    the algorithm returns the graph among $G_0, ..., G_n$ with the highest
    density.

    **Greedy++:**
    Boob et al. [2]_ generalized this algorithm into Greedy++, an iterative
    algorithm that runs several rounds of "peeling". In fact, Greedy++ with 1
    iteration is precisely Charikar's algorithm. The algorithm converges to a
    $(1-\epsilon)$ approximate densest subgraph in $O(\Delta(G)\log
    n/\epsilon^2)$ iterations, where $\Delta(G)$ is the maximum degree,
    and $n$ is the number of nodes in $G$. The algorithm also has other
    desirable properties as shown by [4]_ and [5]_.

    **FISTA Algorithm:**
    Harb et al. [3]_ gave a faster and more scalable algorithm using ideas
    from quadratic programming for the densest subgraph, which is based on a
    fast iterative shrinkage-thresholding algorithm (FISTA) algorithm. It is
    known that computing the densest subgraph can be formulated as the
    following convex optimization problem:

    Minimize $\sum_{u \in V(G)} b_u^2$

    Subject to:

    $b_u = \sum_{v: \{u,v\} \in E(G)} x_{uv}$ for all $u \in V(G)$

    $x_{uv} + x_{vu} = 1.0$ for all $\{u,v\} \in E(G)$

    $x_{uv} \geq 0, x_{vu} \geq 0$ for all $\{u,v\} \in E(G)$

    Here, $x_{uv}$ represents the fraction of edge $\{u,v\}$ assigned to
    $u$, and $x_{vu}$ to $v$.

    The FISTA algorithm efficiently solves this convex program using gradient
    descent with projections. For a learning rate $\alpha$, the algorithm
    does:

    1. **Initialization**: Set $x^{(0)}_{uv} = x^{(0)}_{vu} = 0.5$ for all
    edges as a feasible solution.

    2. **Gradient Update**: For iteration $k\geq 1$, set
    $x^{(k+1)}_{uv} = x^{(k)}_{uv} - 2 \alpha \sum_{v: \{u,v\} \in E(G)}
    x^{(k)}_{uv}$. However, now $x^{(k+1)}_{uv}$ might be infeasible!
    To ensure feasibility, we project $x^{(k+1)}_{uv}$.

    3. **Projection to the Feasible Set**: Compute
    $b^{(k+1)}_u = \sum_{v: \{u,v\} \in E(G)} x^{(k)}_{uv}$ for all
    nodes $u$. Define $z^{(k+1)}_{uv} = x^{(k+1)}_{uv} - 2 \alpha
    b^{(k+1)}_u$. Update $x^{(k+1)}_{uv} =
    CLAMP((z^{(k+1)}_{uv} - z^{(k+1)}_{vu} + 1.0) / 2.0)$, where
    $CLAMP(x) = \max(0, \min(1, x))$.

    With a learning rate of $\alpha=1/\Delta(G)$, where $\Delta(G)$ is
    the maximum degree, the algorithm converges to the optimum solution of
    the convex program.

    **Fractional Peeling:**
    To obtain a **discrete** subgraph, we use fractional peeling, an
    adaptation of the standard peeling algorithm which peels the minimum
    degree vertex in each iteration, and returns the densest subgraph found
    along the way. Here, we instead peel the vertex with the smallest
    induced load $b_u$:

    1. Compute $b_u$ and $x_{uv}$.

    2. Iteratively remove the vertex with the smallest $b_u$, updating its
    neighbors' load by $x_{vu}$.

    Fractional peeling transforms the approximately optimal fractional
    values $b_u, x_{uv}$ into a discrete subgraph. Unlike traditional
    peeling, which removes the lowest-degree node, this method accounts for
    fractional edge contributions from the convex program.

    This approach is both scalable and theoretically sound, ensuring a quick
    approximation of the densest subgraph while leveraging fractional load
    balancing.

    References
    ----------
    .. [1] Charikar, Moses. "Greedy approximation algorithms for finding dense
       components in a graph." In International workshop on approximation
       algorithms for combinatorial optimization, pp. 84-95. Berlin, Heidelberg:
       Springer Berlin Heidelberg, 2000.

    .. [2] Boob, Digvijay, Yu Gao, Richard Peng, Saurabh Sawlani, Charalampos
       Tsourakakis, Di Wang, and Junxing Wang. "Flowless: Extracting densest
       subgraphs without flow computations." In Proceedings of The Web Conference
       2020, pp. 573-583. 2020.

    .. [3] Harb, Elfarouk, Kent Quanrud, and Chandra Chekuri. "Faster and scalable
       algorithms for densest subgraph and decomposition." Advances in Neural
       Information Processing Systems 35 (2022): 26966-26979.

    .. [4] Harb, Elfarouk, Kent Quanrud, and Chandra Chekuri. "Convergence to
       lexicographically optimal base in a (contra) polymatroid and applications
       to densest subgraph and tree packing." arXiv preprint arXiv:2305.02987
       (2023).

    .. [5] Chekuri, Chandra, Kent Quanrud, and Manuel R. Torres. "Densest
       subgraph: Supermodularity, iterative peeling, and flow." In Proceedings of
       the 2022 Annual ACM-SIAM Symposium on Discrete Algorithms (SODA), pp.
       1531-1555. Society for Industrial and Applied Mathematics, 2022.
    """
    try:
        algo = ALGORITHMS[method]
    except KeyError as e:
        raise ValueError(f"{method} is not a valid choice for an algorithm.") from e

    return algo(G, iterations)



r"""Computation of graph non-randomness."""

import math

import networkx as nx
from networkx.utils import not_implemented_for

__all__ = ["non_randomness"]


@not_implemented_for("directed")
@not_implemented_for("multigraph")
@nx._dispatchable(edge_attrs="weight")
def non_randomness(G, k=None, weight="weight"):
    """Compute the non-randomness of a graph.

    The first value $R_G$ is the sum of non-randomness values of all
    edges within the graph (where the non-randomness of an edge tends to be
    small when the two nodes linked by that edge are from two different
    communities).

    The second value $R_G^*$ is a relative measure that indicates
    to what extent `G` is different from a random graph in terms
    of probability. The closer it is to 0, the higher the likelihood
    the graph was generated by an Erdős--Rényi model.

    Parameters
    ----------
    G : NetworkX graph
        Graph must be undirected, connected, and without self-loops.

    k : int or None, optional (default=None)
        The number of communities in `G`.
        If `k` is not set, the function uses a default community detection
        algorithm (:func:`~networkx.algorithms.community.label_propagation_communities`)
        to set it.

    weight : string or None, optional (default="weight")
        The name of an edge attribute that holds the numerical value used
        as a weight. If `None`, then each edge has weight 1, i.e., the graph is
        binary.

    Returns
    -------
    (float, float) tuple
        The first value is $R_G$, the non-randomness of the graph,
        the second is $R_G^*$, the relative non-randomness
        w.r.t. the Erdős--Rényi model.

    Raises
    ------
    NetworkXNotImplemented
        If the input graph is directed or a multigraph.

    NetworkXException
        If the input graph is not connected.

    NetworkXError
        If the input graph contains self-loops or has no edges.

    ValueError
        If `k` is not in $\\{1, \\dots, n-1\\}$, where $n$ is the number of nodes,
        or if `k` is such that the computed edge probability
        $p = \\frac{2km}{n(n-k)}$ does not satisfy $0 < p < 1$.

    Examples
    --------
    >>> G = nx.karate_club_graph()
    >>> nr, nr_rd = nx.non_randomness(G, 2)
    >>> nr, nr_rd = nx.non_randomness(G, 2, "weight")

    When the number of communities `k` is not specified,
    :func:`~networkx.algorithms.community.label_propagation_communities`
    is used to compute it.
    This algorithm can give different results depending on
    the order of nodes and edges in the graph.
    For example, while the following graphs are identical,
    computing the non-randomness of each of them yields different results:

    >>> G1, G2 = nx.Graph(), nx.Graph()
    >>> G1.add_edges_from([(0, 1), (1, 2), (1, 3), (3, 4)])
    >>> G2.add_edges_from([(0, 1), (1, 3), (1, 2), (3, 4)])
    >>> [round(r, 6) for r in nx.non_randomness(G1)]
    [-1.847759, -5.842437]
    >>> [round(r, 6) for r in nx.non_randomness(G2)]
    Traceback (most recent call last):
     ...
    ValueError: invalid number of communities for graph with 5 nodes and 4 edges: 2

    This is because the community detection algorithm finds
    1 community in `G1` and 2 communities in `G2`.
    This can be resolved by specifying the number of communities `k`:

    >>> [round(r, 6) for r in nx.non_randomness(G2, k=1)]
    [-1.847759, -5.842437]

    Notes
    -----
    If a `weight` argument is passed, this algorithm will use the eigenvalues
    of the weighted adjacency matrix instead.

    The output of this function corresponds to (4.4) and (4.5) in [1]_.
    A lower value of $R^*_G$ indicates a more random graph;
    one can think of $1 - \\Phi(R_G^*)$ as the similarity
    between the graph and a random graph,
    where $\\Phi(x)$ is the cumulative distribution function
    of the standard normal distribution.

    Theorem 2 in [2]_ states that for any graph $G$
    with $n$ nodes, $m$ edges, and $k$ communities,
    its non-randomness is bounded below by the non-randomness of an
    $r$-regular graph (a graph where each node has degree $r$),
    and bounded above by the non-randomness of an $l$-complete graph
    (a graph where each community is a clique of $l$ nodes).

    References
    ----------
    .. [1] Xiaowei Ying and Xintao Wu,
           On Randomness Measures for Social Networks,
           SIAM International Conference on Data Mining. 2009
           https://doi.org/10.1137/1.9781611972795.61
    .. [2] Ying, Xiaowei & Wu, Leting & Wu, Xintao. (2012).
           A Spectrum-Based Framework for Quantifying Randomness of Social Networks.
           IEEE Transactions on Knowledge and Data Engineering 23(12):1842--1856.
           https://dl.acm.org/doi/abs/10.1109/TKDE.2010.218
    """
    import numpy as np

    # corner case: graph has no edges
    if nx.is_empty(G):
        raise nx.NetworkXError("non_randomness not applicable to empty graphs")
    if not nx.is_connected(G):
        raise nx.NetworkXException("Non connected graph.")
    if len(list(nx.selfloop_edges(G))) > 0:
        raise nx.NetworkXError("Graph must not contain self-loops")

    n = G.number_of_nodes()
    m = G.number_of_edges()

    if k is None:
        k = len(tuple(nx.community.label_propagation_communities(G)))
    if not 1 <= k < n or not 0 < (p := (2 * k * m) / (n * (n - k))) < 1:
        err = (
            f"invalid number of communities for graph with {n} nodes and {m} edges: {k}"
        )
        raise ValueError(err)

    # eq. 4.4
    eigenvalues = np.linalg.eigvals(nx.to_numpy_array(G, weight=weight))
    nr = float(np.real(np.sum(eigenvalues[:k])))

    # eq. 4.5
    nr_rd = (nr - ((n - 2 * k) * p + k)) / math.sqrt(2 * k * p * (1 - p))

    return nr, nr_rd



import itertools

import networkx as nx

__all__ = ["birank"]


@nx._dispatchable(edge_attrs="weight")
def birank(
    G,
    nodes,
    *,
    alpha=None,
    beta=None,
    top_personalization=None,
    bottom_personalization=None,
    max_iter=100,
    tol=1.0e-6,
    weight="weight",
):
    r"""Compute the BiRank score for nodes in a bipartite network.

    Given the bipartite sets $U$ and $P$, the BiRank algorithm seeks to satisfy
    the following recursive relationships between the scores of nodes $j \in P$
    and $i \in U$:

    .. math::

        p_j = \alpha \sum_{i \in U} \frac{w_{ij}}{\sqrt{d_i}\sqrt{d_j}} u_i
        + (1 - \alpha) p_j^0

        u_i = \beta \sum_{j \in P} \frac{w_{ij}}{\sqrt{d_i}\sqrt{d_j}} p_j
        + (1 - \beta) u_i^0

    where

    * $p_j$ and $u_i$ are the BiRank scores of nodes $j \in P$ and $i \in U$.
    * $w_{ij}$ is the weight of the edge between nodes $i \in U$ and $j \in P$
      (With a value of 0 if no edge exists).
    * $d_i$ and $d_j$ are the weighted degrees of nodes $i \in U$ and $j \in P$,
      respectively.
    * $p_j^0$ and $u_i^0$ are personalization values that can encode a priori
      weights for the nodes $j \in P$ and $i \in U$, respectively. Akin to the
      personalization vector used by PageRank.
    * $\alpha$ and $\beta$ are damping hyperparameters applying to nodes in $P$
      and $U$ respectively. They can take values in the interval $[0, 1]$, and
      are analogous to those used by PageRank.

    Below are two use cases for this algorithm.

    1. Personalized Recommendation System
        Given a bipartite graph representing users and items, BiRank can be used
        as a collaborative filtering algorithm to recommend items to users.
        Previous ratings are encoded as edge weights, and the specific ratings
        of an individual user on a set of items is used as the personalization
        vector over items. See the example below for an implementation of this
        on a toy dataset provided in [1]_.

    2. Popularity Prediction
        Given a bipartite graph representing user interactions with items, e.g.
        commits to a GitHub repository, BiRank can be used to predict the
        popularity of a given item. Edge weights should encode the strength of
        the interaction signal. This could be a raw count, or weighted by a time
        decay function like that specified in Eq. (15) of [1]_. The
        personalization vectors can be used to encode existing popularity
        signals, for example, the monthly download count of a repository's
        package.

    Parameters
    ----------
    G : graph
        A bipartite network

    nodes : iterable of nodes
        Container with all nodes belonging to the first bipartite node set
        ('top'). The nodes in this set use the hyperparameter `alpha`, and the
        personalization dictionary `top_personalization`. The nodes in the second
        bipartite node set ('bottom') are automatically determined by taking the
        complement of 'top' with respect to the graph `G`.

    alpha : float, optional (default=0.80 if top_personalization not empty, else 1)
        Damping factor for the 'top' nodes. Must be in the interval $[0, 1]$.
        Larger alpha and beta generally reduce the effect of the personalizations
        and increase the number of iterations before convergence. Choice of value
        is largely dependent on use case, and experimentation is recommended.

    beta : float, optional (default=0.80 if bottom_personalization not empty, else 1)
        Damping factor for the 'bottom' nodes. Must be in the interval $[0, 1]$.
        Larger alpha and beta generally reduce the effect of the personalizations
        and increase the number of iterations before convergence. Choice of value
        is largely dependent on use case, and experimentation is recommended.

    top_personalization : dict, optional (default=None)
        Dictionary keyed by nodes in 'top' to that node's personalization value.
        Unspecified nodes in 'top' will be assigned a personalization value of 0.
        Personalization values are used to encode a priori weights for a given node,
        and should be non-negative.

    bottom_personalization : dict, optional (default=None)
        Dictionary keyed by nodes in 'bottom' to that node's personalization value.
        Unspecified nodes in 'bottom' will be assigned a personalization value of 0.
        Personalization values are used to encode a priori weights for a given node,
        and should be non-negative.

    max_iter : int, optional (default=100)
        Maximum number of iterations in power method eigenvalue solver.

    tol : float, optional (default=1.0e-6)
        Error tolerance used to check convergence in power method solver. The
        iteration will stop after a tolerance of both ``len(top) * tol`` and
        ``len(bottom) * tol`` is reached for nodes in 'top' and 'bottom'
        respectively.

    weight : string or None, optional (default='weight')
        Edge data key to use as weight.

    Returns
    -------
    birank : dictionary
        Dictionary keyed by node to that node's BiRank score.

    Raises
    ------
    NetworkXAlgorithmError
        If the parameters `alpha` or `beta` are not in the interval [0, 1],
        if either of the bipartite sets are empty, or if negative values are
        provided in the personalization dictionaries.

    PowerIterationFailedConvergence
        If the algorithm fails to converge to the specified tolerance
        within the specified number of iterations of the power iteration
        method.

    Examples
    --------
    Construct a bipartite graph with user-item ratings and use BiRank to
    recommend items to a user (user 1). The example below uses the `rating`
    edge attribute as the weight of the edges. The `top_personalization` vector
    is used to encode the user's previous ratings on items.

    Creation of graph, bipartite sets for the example.

    >>> elist = [
    ...     ("u1", "p1", 5),
    ...     ("u2", "p1", 5),
    ...     ("u2", "p2", 4),
    ...     ("u3", "p1", 3),
    ...     ("u3", "p3", 2),
    ... ]
    >>> G = nx.Graph()
    >>> G.add_weighted_edges_from(elist, weight="rating")
    >>> product_nodes = ("p1", "p2", "p3")
    >>> user = "u1"

    First, we create a personalization vector for the user based on on their
    ratings of past items. In this case they have only rated one item (p1, with
    a rating of 5) in the past.

    >>> user_personalization = {
    ...     product: rating
    ...     for _, product, rating in G.edges(nbunch=user, data="rating")
    ... }
    >>> user_personalization
    {'p1': 5}

    Calculate the BiRank score of all nodes in the graph, filter for the items
    that the user has not rated yet, and sort the results by score.

    >>> user_birank_results = nx.bipartite.birank(
    ...     G, product_nodes, top_personalization=user_personalization, weight="rating"
    ... )
    >>> user_birank_results = filter(
    ...     lambda item: item[0][0] == "p" and user not in G.neighbors(item[0]),
    ...     user_birank_results.items(),
    ... )
    >>> user_birank_results = sorted(
    ...     user_birank_results, key=lambda item: item[1], reverse=True
    ... )
    >>> user_recommendations = {
    ...     product: round(score, 5) for product, score in user_birank_results
    ... }
    >>> user_recommendations
    {'p2': 1.44818, 'p3': 1.04811}

    We find that user 1 should be recommended item p2 over item p3. This is due
    to the fact that user 2 rated also rated p1 highly, while user 3 did not.
    Thus user 2's tastes are inferred to be similar to user 1's, and carry more
    weight in the recommendation.

    See Also
    --------
    :func:`~networkx.algorithms.link_analysis.pagerank_alg.pagerank`
    :func:`~networkx.algorithms.link_analysis.hits_alg.hits`
    :func:`~networkx.algorithms.bipartite.centrality.betweenness_centrality`
    :func:`~networkx.algorithms.bipartite.basic.sets`
    :func:`~networkx.algorithms.bipartite.basic.is_bipartite`

    Notes
    -----
    The `nodes` input parameter must contain all nodes in one bipartite
    node set, but the dictionary returned contains all nodes from both
    bipartite node sets. See :mod:`bipartite documentation
    <networkx.algorithms.bipartite>` for further details on how
    bipartite graphs are handled in NetworkX.

    In the case a personalization dictionary is not provided for top (bottom)
    `alpha` (`beta`) will default to 1. This is because a damping factor
    without a non-zero entry in the personalization vector will lead to the
    algorithm converging to the zero vector.

    References
    ----------
    .. [1] Xiangnan He, Ming Gao, Min-Yen Kan, and Dingxian Wang. 2017.
       BiRank: Towards Ranking on Bipartite Graphs. IEEE Trans. on Knowl.
       and Data Eng. 29, 1 (January 2017), 57–71.
       https://arxiv.org/pdf/1708.04396

    """
    import numpy as np
    import scipy as sp

    # Initialize the sets of top and bottom nodes
    top = set(nodes)
    bottom = set(G) - top
    top_count = len(top)
    bottom_count = len(bottom)

    if top_count == 0 or bottom_count == 0:
        raise nx.NetworkXAlgorithmError(
            "The BiRank algorithm requires a bipartite graph with at least one"
            "node in each set."
        )

    # Clean the personalization dictionaries
    top_personalization = _clean_personalization_dict(top_personalization)
    bottom_personalization = _clean_personalization_dict(bottom_personalization)

    # Set default values for alpha and beta if not provided
    if alpha is None:
        alpha = 0.8 if top_personalization else 1
    if beta is None:
        beta = 0.8 if bottom_personalization else 1

    if alpha < 0 or alpha > 1:
        raise nx.NetworkXAlgorithmError("alpha must be in the interval [0, 1]")
    if beta < 0 or beta > 1:
        raise nx.NetworkXAlgorithmError("beta must be in the interval [0, 1]")

    # Initialize query vectors
    p0 = np.array([top_personalization.get(n, 0) for n in top], dtype=float)
    u0 = np.array([bottom_personalization.get(n, 0) for n in bottom], dtype=float)

    # Construct degree normalized biadjacency matrix `S` and its transpose
    W = nx.bipartite.biadjacency_matrix(G, bottom, top, weight=weight, dtype=float)
    p_degrees = W.sum(axis=0, dtype=float)
    # Handle case where the node is disconnected - avoids warning
    p_degrees[p_degrees == 0] = 1.0
    D_p = sp.sparse.dia_array(
        ([1.0 / np.sqrt(p_degrees)], [0]),
        shape=(top_count, top_count),
        dtype=float,
    )
    u_degrees = W.sum(axis=1, dtype=float)
    u_degrees[u_degrees == 0] = 1.0
    D_u = sp.sparse.dia_array(
        ([1.0 / np.sqrt(u_degrees)], [0]),
        shape=(bottom_count, bottom_count),
        dtype=float,
    )
    S = D_u.tocsr() @ W @ D_p.tocsr()
    S_T = S.T

    # Initialize birank vectors for iteration
    p = np.ones(top_count, dtype=float) / top_count
    u = beta * (S @ p) + (1 - beta) * u0

    # Iterate until convergence
    for _ in range(max_iter):
        p_last = p
        u_last = u
        p = alpha * (S_T @ u) + (1 - alpha) * p0
        u = beta * (S @ p) + (1 - beta) * u0

        # Continue iterating if the error (absolute if less than 1, relative otherwise)
        # is above the tolerance threshold for either p or u
        err_u = np.absolute((u_last - u) / np.maximum(1.0, u_last)).sum()
        if err_u >= len(u) * tol:
            continue
        err_p = np.absolute((p_last - p) / np.maximum(1.0, p_last)).sum()
        if err_p >= len(p) * tol:
            continue

        # Handle edge case where if both alpha and beta are 1, scale is
        # indeterminate, so normalization is required to return consistent results
        if alpha == 1 and beta == 1:
            p = p / np.linalg.norm(p, 1)
            u = u / np.linalg.norm(u, 1)

        # If both error thresholds pass, return a single dictionary mapping
        # nodes to their scores
        return dict(
            zip(itertools.chain(top, bottom), map(float, itertools.chain(p, u)))
        )

    # If we reach this point, we have not converged
    raise nx.PowerIterationFailedConvergence(max_iter)


def _clean_personalization_dict(personalization):
    """Filter out zero values from the personalization dictionary,
    handle case where None is passed, ensure values are non-negative."""
    if personalization is None:
        return {}
    if any(value < 0 for value in personalization.values()):
        raise nx.NetworkXAlgorithmError("Personalization values must be non-negative.")
    return {node: value for node, value in personalization.items() if value != 0}



"""Functions for splitting a network into two communities (finding a bipartition)."""

import random
from copy import deepcopy
from itertools import count

import networkx as nx

__all__ = [
    "kernighan_lin_bisection",
    "spectral_modularity_bipartition",
    "greedy_node_swap_bipartition",
]


def _kernighan_lin_sweep(edge_info, side):
    """
    This is a modified form of Kernighan-Lin, which moves single nodes at a
    time, alternating between sides to keep the bisection balanced.  We keep
    two min-heaps of swap costs to make optimal-next-move selection fast.
    """
    heap0, heap1 = cost_heaps = nx.utils.BinaryHeap(), nx.utils.BinaryHeap()
    # we use heap methods insert, pop, and get
    for u, nbrs in edge_info.items():
        cost_u = sum(wt if side[v] else -wt for v, wt in nbrs.items())
        if side[u]:
            heap1.insert(u, cost_u)
        else:
            heap0.insert(u, -cost_u)

    def _update_heap_values(node):
        side_node = side[node]
        for nbr, wt in edge_info[node].items():
            side_nbr = side[nbr]
            if side_nbr == side_node:
                wt = -wt
            heap_nbr = cost_heaps[side_nbr]
            if nbr in heap_nbr:
                cost_nbr = heap_nbr.get(nbr) + 2 * wt
                # allow_increase lets us update a value already on the heap
                heap_nbr.insert(nbr, cost_nbr, allow_increase=True)

    i = 0
    totcost = 0
    while heap0 and heap1:
        u, cost_u = heap0.pop()
        _update_heap_values(u)
        v, cost_v = heap1.pop()
        _update_heap_values(v)
        totcost += cost_u + cost_v
        i += 1
        yield totcost, i, (u, v)


@nx.utils.not_implemented_for("directed")
@nx.utils.py_random_state(4)
@nx._dispatchable(edge_attrs="weight")
def kernighan_lin_bisection(G, partition=None, max_iter=10, weight="weight", seed=None):
    """Partition a graph into two blocks using the Kernighan–Lin algorithm.

    This algorithm partitions a network into two sets by iteratively
    swapping pairs of nodes to reduce the edge cut between the two sets.  The
    pairs are chosen according to a modified form of Kernighan-Lin [1]_, which
    moves nodes individually, alternating between sides to keep the bisection
    balanced.

    Kernighan-Lin is an approximate algorithm for maximal modularity bisection.
    In [2]_ they suggest that fine-tuned improvements can be made using
    greedy node swapping, (see `greedy_node_swap_bipartition`).
    The improvements are typically only a few percent of the modularity value.
    But they claim that can make a difference between a good and excellent method.
    This function does not perform any improvements. But you can do that yourself.

    Parameters
    ----------
    G : NetworkX graph
        Graph must be undirected.

    partition : tuple
        Pair of iterables containing an initial partition. If not
        specified, a random balanced partition is used.

    max_iter : int
        Maximum number of times to attempt swaps to find an
        improvement before giving up.

    weight : string or function (default: "weight")
        If this is a string, then edge weights will be accessed via the
        edge attribute with this key (that is, the weight of the edge
        joining `u` to `v` will be ``G.edges[u, v][weight]``). If no
        such edge attribute exists, the weight of the edge is assumed to
        be one.

        If this is a function, the weight of an edge is the value
        returned by the function. The function must accept exactly three
        positional arguments: the two endpoints of an edge and the
        dictionary of edge attributes for that edge. The function must
        return a number or None to indicate a hidden edge.

    seed : integer, random_state, or None (default)
        Indicator of random number generation state.
        See :ref:`Randomness<randomness>`.
        Only used if partition is None

    Returns
    -------
    partition : tuple
        A pair of sets of nodes representing the bipartition.

    Raises
    ------
    NetworkXError
        If `partition` is not a valid partition of the nodes of the graph.

    References
    ----------
    .. [1] Kernighan, B. W.; Lin, Shen (1970).
       "An efficient heuristic procedure for partitioning graphs."
       *Bell Systems Technical Journal* 49: 291--307.
       Oxford University Press 2011.
    .. [2] M. E. J. Newman,
       "Modularity and community structure in networks",
       PNAS, 103 (23), p. 8577-8582,
       https://doi.org/10.1073/pnas.0601602103

    """
    nodes = list(G)

    if partition is None:
        seed.shuffle(nodes)
        mid = len(nodes) // 2
        A, B = nodes[:mid], nodes[mid:]
    else:
        try:
            A, B = partition
        except (TypeError, ValueError) as err:
            raise nx.NetworkXError("partition must be two sets") from err
        if not nx.community.is_partition(G, [A, B]):
            raise nx.NetworkXError("partition invalid")

    side = {node: (node in A) for node in nodes}

    # ruff: noqa: E731   skips check for no lambda functions
    # Using shortest_paths _weight_function with sum instead of min on multiedges
    if callable(weight):
        sum_weight = weight
    elif G.is_multigraph():
        sum_weight = lambda u, v, d: sum(dd.get(weight, 1) for dd in d.values())
    else:
        sum_weight = lambda u, v, d: d.get(weight, 1)

    edge_info = {
        u: {v: wt for v, d in nbrs.items() if (wt := sum_weight(u, v, d)) is not None}
        for u, nbrs in G._adj.items()
    }

    for i in range(max_iter):
        costs = list(_kernighan_lin_sweep(edge_info, side))
        # find out how many edges to update: min_i
        min_cost, min_i, _ = min(costs)
        if min_cost >= 0:
            break

        for _, _, (u, v) in costs[:min_i]:
            side[u] = 1
            side[v] = 0

    part1 = {u for u, s in side.items() if s == 0}
    part2 = {u for u, s in side.items() if s == 1}
    return part1, part2


@nx.utils.not_implemented_for("directed")
@nx.utils.not_implemented_for("multigraph")
def spectral_modularity_bipartition(G):
    r"""Return a bipartition of the nodes based on the spectrum of the
    modularity matrix of the graph.

    This method calculates the eigenvector associated with the second
    largest eigenvalue of the modularity matrix, where the modularity
    matrix *B* is defined by

    ..math::

        B_{i j} = A_{i j} - \frac{k_i k_j}{2 m},

    where *A* is the adjacency matrix, `k_i` is the degree of node *i*,
    and *m* is the number of edges in the graph. Nodes whose
    corresponding values in the eigenvector are negative are placed in
    one block, nodes whose values are nonnegative are placed in another
    block.

    Parameters
    ----------
    G : NetworkX graph

    Returns
    -------
    C : tuple
        Pair of communities as two sets of nodes of ``G``, partitioned
        according to second largest eigenvalue of the modularity matrix.

    Examples
    --------
    >>> G = nx.karate_club_graph()
    >>> MrHi, Officer = nx.community.spectral_modularity_bipartition(G)
    >>> MrHi, Officer = sorted([sorted(MrHi), sorted(Officer)])
    >>> MrHi
    [0, 1, 2, 3, 4, 5, 6, 7, 10, 11, 12, 13, 16, 17, 19, 21]
    >>> Officer
    [8, 9, 14, 15, 18, 20, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33]

    References
    ----------
    .. [1] M. E. J. Newman *Networks: An Introduction*, pages 373--378
       Oxford University Press 2011.
    .. [2] M. E. J. Newman,
       "Modularity and community structure in networks",
       PNAS, 103 (23), p. 8577-8582,
       https://doi.org/10.1073/pnas.0601602103

    """
    import numpy as np

    B = nx.linalg.modularity_matrix(G)
    eigenvalues, eigenvectors = np.linalg.eig(B)
    index = np.argsort(eigenvalues)[-1]
    v2 = zip(np.real(eigenvectors[:, index]), G)
    left, right = set(), set()
    for u, n in v2:
        if u < 0:
            left.add(n)
        else:
            right.add(n)
    return left, right


@nx.utils.not_implemented_for("multigraph")
def greedy_node_swap_bipartition(G, *, init_split=None, max_iter=10):
    """Split the nodes into two communities based on greedy
    modularity maximization.

    The algorithm works by selecting a node to change communities which
    will maximize the modularity. The swap is made and the community
    structure with the highest modularity is kept.

    Parameters
    ----------
    G : NetworkX graph

    init_split : 2-tuple of sets of nodes
        Pair of sets of nodes in ``G`` providing an initial bipartition
        for the algorithm. If not specified, a random balanced partition
        is used. If this pair of sets is not a partition of the nodes of `G`,
        :exc:`NetworkXException` is raised.

    max_iter : int
      Maximum number of iterations of attempting swaps to find an improvement.

    Returns
    -------
    max_split : 2-tuple of sets of nodes
        Pair of sets of nodes of ``G``, partitioned according to a
        node swap greedy modularity maximization algorithm.

    Raises
    ------
    NetworkXError
      if init_split is not a valid partition of the
      graph into two communities or if G is a MultiGraph

    Examples
    --------
    >>> G = nx.barbell_graph(3, 0)
    >>> left, right = nx.community.greedy_node_swap_bipartition(G)
    >>> # Sort the communities so the nodes appear in increasing order.
    >>> left, right = sorted([sorted(left), sorted(right)])
    >>> sorted(left)
    [0, 1, 2]
    >>> sorted(right)
    [3, 4, 5]

    Notes
    -----
    This function is not implemented for multigraphs.

    References
    ----------
    .. [1] M. E. J. Newman "Networks: An Introduction", pages 373--375.
       Oxford University Press 2011.

    """
    if init_split is None:
        m1 = len(G) // 2
        m2 = len(G) - m1
        some_nodes = set(random.sample(list(G), m1))
        other_nodes = {n for n in G if n not in some_nodes}
        best_split_so_far = (some_nodes, other_nodes)
    else:
        if not nx.community.is_partition(G, init_split):
            raise nx.NetworkXError("init_split is not a partition of G")
        if not len(init_split) == 2:
            raise nx.NetworkXError("init_split must be a bipartition")
        best_split_so_far = deepcopy(init_split)

    best_mod = nx.community.modularity(G, best_split_so_far)

    max_split, max_mod = best_split_so_far, best_mod
    its = 0
    m = G.number_of_edges()
    G_degree = dict(G.degree)

    while max_mod >= best_mod and its < max_iter:
        best_split_so_far = max_split
        best_mod = max_mod
        next_split = deepcopy(max_split)
        next_mod = max_mod
        nodes = set(G)
        while nodes:
            max_swap = -1
            max_node = None
            max_node_comm = None
            left, right = next_split
            leftd = sum(G_degree[n] for n in left)
            rightd = sum(G_degree[n] for n in right)
            for n in nodes:
                if n in left:
                    in_comm, out_comm = left, right
                    in_deg, out_deg = leftd, rightd
                else:
                    in_comm, out_comm = right, left
                    in_deg, out_deg = rightd, leftd

                d_eii = -len(G[n].keys() & in_comm) / m
                d_ejj = len(G[n].keys() & out_comm) / m
                deg = G_degree[n]
                d_sum_ai = (deg / (2 * m**2)) * (in_deg - out_deg - deg)
                swap_change = d_eii + d_ejj + d_sum_ai

                if swap_change > max_swap:
                    max_swap = swap_change
                    max_node = n
                    max_node_comm = in_comm
                    non_max_node_comm = out_comm
            # swap the node from one comm to the other
            max_node_comm.remove(max_node)
            non_max_node_comm.add(max_node)
            next_mod += max_swap
            # deepcopy next_split each time it reaches a high (might go lower later)
            if next_mod > max_mod:
                max_split, max_mod = deepcopy(next_split), next_mod
            nodes.remove(max_node)
        its += 1
    return best_split_so_far



"""
===============
Cycle Detection
===============

This example demonstrates the use of ``nx.find_cycle`` to find a single,
arbitrary cycle in a graph.

Other functions like ``nx.simple_cycles`` and ``nx.cycle_basis`` can be used to
find all cycles or a cycle basis.
"""

import networkx as nx
import matplotlib.pyplot as plt

# Create a simple directed graph with a cycle
G = nx.DiGraph([(1, 2), (2, 3), (3, 4), (4, 2), (3, 5), (3, 2), (1, 5)])

# Draw the graph
pos = nx.spring_layout(G, seed=8020)
nx.draw(G, pos, with_labels=True)

# The `orientation` parameter can be used to determine how directed edges are
# treated and the reporting of edge direction in the cycle
cycle = nx.find_cycle(G, orientation="original")
print(cycle)

# Highlight the cycle in red
nx.draw_networkx_edges(G, pos, edgelist=cycle, edge_color="r", width=2)
plt.show()



"""
==================================================
Image Segmentation via Spectral Graph Partitioning
==================================================

Example of partitioning a undirected graph obtained by ``k-neighbors``
from an RGB image into two subgraphs using spectral clustering
illustrated by 3D plots of the original labeled data points in RGB 3D space
vs the bi-partition marking performed by graph partitioning via spectral clustering.
All 3D plots use the 3D spectral layout.

See :ref:`sphx_glr_auto_examples_3d_drawing` for recipes to create 3D animations
from these visualizations.
"""

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.lines import Line2D
from sklearn.cluster import SpectralClustering

# sphinx_gallery_thumbnail_number = 3

###############################################################################
# Create an example 3D dataset "The Rings".
# -----------------------------------------
# The dataset is made of two entangled noisy rings in 3D.
np.random.seed(0)
N_SAMPLES = 128
X = np.random.random((N_SAMPLES, 3)) * 5e-1
m = int(np.round(N_SAMPLES / 2))
theta = np.linspace(0, 2 * np.pi, m)
X[0:m, 0] += 2 * np.cos(theta)
X[0:m, 1] += 3 * np.sin(theta) + 1
X[0:m, 2] += np.sin(theta) + 0.5
X[m:, 0] += 2 * np.sin(theta)
X[m:, 1] += 2 * np.cos(theta) - 1
X[m:, 2] += 3 * np.sin(theta)
Y = np.zeros(N_SAMPLES, dtype=np.int8)
Y[m:] = np.ones(m, dtype=np.int8)

# map X to int8 for 8-bit RGB interpretation for drawing
for i in np.arange(X.shape[1]):
    x = X[:, i]
    min_x = np.min(x)
    max_x = np.max(x)
    X[:, i] = np.round(255 * (x - min_x) / (max_x - min_x))

###############################################################################
# Plot the RGB dataset as an image.
# ---------------------------------
# Each of the 128 3D data points can be treated as RGB values of a pixel.
# We illustrate the dataset plotting it as an image with 8x16 pixels,
# with the pixels randomly placed in the 8x16 grid.
# The fine structure in the data is not visually detectable in the image.

perm = np.random.permutation(X.shape[0])
rgb_array = X[perm, :].reshape(8, 16, 3).astype(int)
fig, ax = plt.subplots()
ax.matshow(rgb_array)
plt.show()

###############################################################################
# Generate the graph and determine the two clusters.
# --------------------------------------------------
# The graph is constructed using the "nearest_neighbors" and the two clusters
# are determined by spectral clustering/graph partitioning.
NUM_CLUSTERS = 2
sc = SpectralClustering(
    n_clusters=NUM_CLUSTERS,
    affinity="nearest_neighbors",
    random_state=4242,
    n_neighbors=10,
    assign_labels="cluster_qr",
    n_jobs=-1,
)
clusters = sc.fit(X)
cluster_affinity_matrix = clusters.affinity_matrix_.getH()

pred_labels = clusters.labels_.astype(int)
G = nx.from_scipy_sparse_array(cluster_affinity_matrix)
# remove self edges
G.remove_edges_from(nx.selfloop_edges(G))

cluser_member = []
for u in G.nodes:
    cluser_member.append(pred_labels[u])

###############################################################################
# Generate the plots of the data.
# -------------------------------
# The data points are marked according to the original labels (left panel)
# and via clustering (right panel).


def _scatter_plot(ax, X, array_of_markers, axis_plot=True):
    # `marker` parameter does not support list or array format, needs a loop
    for i, marker in enumerate(array_of_markers):
        ax.scatter(
            X[i, 0],
            X[i, 1],
            X[i, 2],
            s=26,
            marker=marker,
            alpha=0.8,
            color=tuple(X[i] / 255),
        )
    if axis_plot:
        ax.set_xlabel("Red")
        ax.set_ylabel("Green")
        ax.set_zlabel("Blue")
    else:
        ax.set_axis_off()
    ax.grid(False)
    ax.view_init(elev=6.0, azim=-22.0)


# select the second half of the list of markers for better visibility
list_of_markers = Line2D.filled_markers[len(Line2D.filled_markers) // 2 :]

fig = plt.figure(figsize=(10, 5))
fig.suptitle("Spectral Clustering as Graph Partitioning Illustrated", fontsize=20)

ax0 = fig.add_subplot(1, 2, 1, projection="3d")
ax0.set_title("Original labeled RGB data")
array_of_markers = np.array(list_of_markers)[Y.astype(int)]
_scatter_plot(ax0, X, array_of_markers)

ax1 = fig.add_subplot(1, 2, 2, projection="3d")
ax1.set_title("Data marked by clustering")
array_of_markers = np.array(list_of_markers)[pred_labels.astype(int)]
_scatter_plot(ax1, X, array_of_markers)

plt.show()


###############################################################################
# Generate the plots of the graph.
# --------------------------------
# The nodes of the graph are marked according to clustering.

# get affinity matrix from spectral clustering
weights = [d["weight"] for u, v, d in G.edges(data=True)]

fig = plt.figure(figsize=(10, 5))
ax0 = fig.add_subplot(1, 2, 1)
ax0.set_title("Graph of Affinity Matrix by k-neighbors in spectral layout")
pos = nx.spectral_layout(G)
nx.draw_networkx(
    G,
    pos=pos,
    alpha=0.5,
    node_size=50,
    with_labels=False,
    ax=ax0,
    node_color=X / 255,
    edge_color="Grey",
)
plt.box(False)
ax0.grid(False)
ax0.set_axis_off()

ax1 = fig.add_subplot(1, 2, 2, projection="3d")
ax1.set_title("Partitioned graph by spectral clustering")
pos = nx.spectral_layout(G, dim=3)
nodes = np.array([pos[v] for v in G])
edges = np.array([(pos[u], pos[v]) for u, v in G.edges()])
point_size = int(800 / np.sqrt(len(nodes)))


def _3d_graph_plot(ax):
    for i, marker in enumerate(array_of_markers):
        ax.scatter(
            *nodes[i].T,
            s=point_size,
            color=tuple(X[i] / 255),
            marker=marker,
            alpha=0.5,
        )
    for vizedge, weight in zip(edges, weights):
        ax.plot(*vizedge.T, color="tab:gray", linewidth=weight, alpha=weight)
    ax.view_init(elev=100.0, azim=-100.0)
    ax.grid(False)
    ax.set_axis_off()


_3d_graph_plot(ax1)
plt.tight_layout()
plt.show()



"""Unit tests for matplotlib drawing functions."""

import itertools
import os
import warnings

import pytest

import networkx as nx

mpl = pytest.importorskip("matplotlib")
np = pytest.importorskip("numpy")
mpl.use("PS")
plt = pytest.importorskip("matplotlib.pyplot")
plt.rcParams["text.usetex"] = False


barbell = nx.barbell_graph(4, 6)

defaults = {
    "node_pos": None,
    "node_visible": True,
    "node_color": "#1f78b4",
    "node_size": 300,
    "node_label": {
        "size": 12,
        "color": "#000000",
        "family": "sans-serif",
        "weight": "normal",
        "alpha": 1.0,
        "background_color": None,
        "background_alpha": None,
        "h_align": "center",
        "v_align": "center",
        "bbox": None,
    },
    "node_shape": "o",
    "node_alpha": 1.0,
    "node_border_width": 1.0,
    "node_border_color": "face",
    "edge_visible": True,
    "edge_width": 1.0,
    "edge_color": "#000000",
    "edge_label": {
        "size": 12,
        "color": "#000000",
        "family": "sans-serif",
        "weight": "normal",
        "alpha": 1.0,
        "bbox": {"boxstyle": "round", "ec": (1.0, 1.0, 1.0), "fc": (1.0, 1.0, 1.0)},
        "h_align": "center",
        "v_align": "center",
        "pos": 0.5,
        "rotate": True,
    },
    "edge_style": "-",
    "edge_alpha": 1.0,
    # These are for undirected-graphs. Directed graphs shouls use "-|>" and 10, respectively
    "edge_arrowstyle": "-",
    "edge_arrowsize": 0,
    "edge_curvature": "arc3",
    "edge_source_margin": 0,
    "edge_target_margin": 0,
}


@pytest.mark.parametrize(
    ("param_name", "param_value", "expected"),
    (
        ("node_color", None, defaults["node_color"]),
        ("node_color", "#FF0000", "red"),
        ("node_color", "color", "lime"),
    ),
)
def test_display_arg_handling_node_color(param_name, param_value, expected):
    G = nx.path_graph(4)
    nx.set_node_attributes(G, "#00FF00", "color")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas, **{param_name: param_value})
    assert mpl.colors.same_color(canvas.get_children()[0].get_edgecolors()[0], expected)
    plt.close()


@pytest.mark.parametrize(
    ("param_value", "expected"),
    (
        (None, (1, 1, 1, 1)),  # default value
        (0.5, (0.5, 0.5, 0.5, 0.5)),
        ("n_alpha", (1.0, 1 / 2, 1 / 3, 0.25)),
    ),
)
def test_display_arg_handling_node_alpha(param_value, expected):
    G = nx.path_graph(4)
    nx.set_node_attributes(G, {n: 1 / (n + 1) for n in G.nodes()}, "n_alpha")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas, node_alpha=param_value)
    assert all(
        canvas.get_children()[0].get_fc()[:, 3] == expected
    )  # Extract just the alpha from the node colors
    plt.close()


def test_display_node_position():
    G = nx.path_graph(4)
    nx.set_node_attributes(G, {n: (n, n) for n in G.nodes()}, "pos")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas, node_pos="pos")
    assert np.all(
        canvas.get_children()[0].get_offsets().data == [[0, 0], [1, 1], [2, 2], [3, 3]]
    )
    plt.close()


def test_display_line_collection():
    G = nx.karate_club_graph()
    nx.set_edge_attributes(
        G, {(u, v): "-|>" if (u + v) % 2 else "-" for u, v in G.edges()}, "arrowstyle"
    )
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas, edge_arrowsize=10)
    # There should only be one line collection in any given visualization
    lc = [
        l
        for l in canvas.get_children()
        if isinstance(l, mpl.collections.LineCollection)
    ][0]
    assert len(lc.get_paths()) == sum([1 for u, v in G.edges() if (u + v) % 2])
    plt.close()


@pytest.mark.parametrize(
    ("edge_color", "expected"),
    (
        (None, "black"),
        ("r", "red"),
        ((1.0, 1.0, 0.0), "yellow"),
        ((0, 1, 0, 1), "lime"),
        ("color", "blue"),
        ("#0000FF", "blue"),
    ),
)
@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_single_color(edge_color, expected, graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_edge_attributes(G, "#0000FF", "color")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, edge_color=edge_color, canvas=canvas)
    if G.is_directed():
        colors = [
            f.get_fc()
            for f in canvas.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        colors = [
            l
            for l in canvas.collections
            if isinstance(l, mpl.collections.LineCollection)
        ][0].get_colors()
    assert all(mpl.colors.same_color(c, expected) for c in colors)
    plt.close()


@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_multiple_colors(graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_edge_attributes(G, {(0, 1): "#FF0000", (1, 2): (0, 0, 1)}, "color")
    ax = plt.figure().add_subplot(111)
    nx.display(G, canvas=ax)
    expected = ["red", "blue"]
    if G.is_directed():
        colors = [
            f.get_fc()
            for f in ax.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        colors = [
            l for l in ax.collections if isinstance(l, mpl.collections.LineCollection)
        ][0].get_colors()
    assert mpl.colors.same_color(colors, expected)
    plt.close()


@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_position(graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_node_attributes(G, {n: (n, n) for n in G.nodes()}, "pos")
    ax = plt.figure().add_subplot(111)
    nx.display(G, canvas=ax)
    if G.is_directed():
        end_points = [
            (f.get_path().vertices[0, :], f.get_path().vertices[-2, :])
            for f in ax.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        line_collection = [
            l for l in ax.collections if isinstance(l, mpl.collections.LineCollection)
        ][0]
        end_points = [
            (p.vertices[0, :], p.vertices[-1, :]) for p in line_collection.get_paths()
        ]
    expected = [((0, 0), (1, 1)), ((1, 1), (2, 2))]
    # Use the threshold to account for slight shifts in FancyArrowPatch margins to
    # avoid covering the arrow head with the node.
    threshold = 0.05
    for a, e in zip(end_points, expected):
        act_start, act_end = a
        exp_start, exp_end = e
        assert all(abs(act_start - exp_start) < (threshold, threshold)) and all(
            abs(act_end - exp_end) < (threshold, threshold)
        )
    plt.close()


def test_display_position_function():
    G = nx.karate_club_graph()

    def fixed_layout(G):
        return nx.spring_layout(G, seed=314159)

    pos = fixed_layout(G)
    ax = plt.figure().add_subplot(111)
    nx.display(G, node_pos=fixed_layout, canvas=ax)
    # rebuild the position dictionary from the canvas
    act_pos = {
        n: tuple(p) for n, p in zip(G.nodes(), ax.get_children()[0].get_offsets().data)
    }
    for n in G.nodes():
        assert all(pos[n] == act_pos[n])
    plt.close()


@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_colormaps(graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_edge_attributes(G, {(0, 1): 0, (1, 2): 1}, "weight")
    cmap = mpl.colormaps["plasma"]
    nx.apply_matplotlib_colors(G, "weight", "color", cmap, nodes=False)
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas)
    mapper = mpl.cm.ScalarMappable(cmap=cmap)
    mapper.set_clim(0, 1)
    expected = [mapper.to_rgba(0), mapper.to_rgba(1)]
    if G.is_directed():
        colors = [
            f.get_facecolor()
            for f in canvas.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        colors = [
            l
            for l in canvas.collections
            if isinstance(l, mpl.collections.LineCollection)
        ][0].get_colors()
    assert mpl.colors.same_color(expected[0], G.edges[0, 1]["color"])
    assert mpl.colors.same_color(expected[1], G.edges[1, 2]["color"])
    assert mpl.colors.same_color(expected, colors)
    plt.close()


@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_node_colormaps(graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_node_attributes(G, {0: 0, 1: 0.5, 2: 1}, "weight")
    cmap = mpl.colormaps["plasma"]
    nx.apply_matplotlib_colors(G, "weight", "color", cmap)
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas)
    mapper = mpl.cm.ScalarMappable(cmap=cmap)
    mapper.set_clim(0, 1)
    expected = [mapper.to_rgba(0), mapper.to_rgba(0.5), mapper.to_rgba(1)]
    colors = [
        s for s in canvas.collections if isinstance(s, mpl.collections.PathCollection)
    ][0].get_edgecolors()
    assert mpl.colors.same_color(expected[0], G.nodes[0]["color"])
    assert mpl.colors.same_color(expected[1], G.nodes[1]["color"])
    assert mpl.colors.same_color(expected[2], G.nodes[2]["color"])
    assert mpl.colors.same_color(expected, colors)
    plt.close()


@pytest.mark.parametrize(
    ("param_value", "expected"),
    (
        (None, [defaults["edge_width"], defaults["edge_width"]]),
        (5, [5, 5]),
        ("width", [5, 10]),
    ),
)
@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_width(param_value, expected, graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_edge_attributes(G, {(0, 1): 5, (1, 2): 10}, "width")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, edge_width=param_value, canvas=canvas)
    if G.is_directed():
        widths = [
            f.get_linewidth()
            for f in canvas.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        widths = list(
            [
                l
                for l in canvas.collections
                if isinstance(l, mpl.collections.LineCollection)
            ][0].get_linewidths()
        )
    assert widths == expected


@pytest.mark.parametrize(
    ("param_value", "expected"),
    (
        (None, [defaults["edge_style"], defaults["edge_style"]]),
        (":", [":", ":"]),
        ("style", ["-", ":"]),
    ),
)
@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_display_edge_style(param_value, expected, graph_type):
    G = nx.path_graph(3, create_using=graph_type)
    nx.set_edge_attributes(G, {(0, 1): "-", (1, 2): ":"}, "style")
    canvas = plt.figure().add_subplot(111)
    nx.display(G, edge_style=param_value, canvas=canvas)
    if G.is_directed():
        styles = [
            f.get_linestyle()
            for f in canvas.get_children()
            if isinstance(f, mpl.patches.FancyArrowPatch)
        ]
    else:
        # Convert back from tuple description to character form
        linestyles = {(0, None): "-", (0, (1, 1.65)): ":"}
        styles = [
            linestyles[(s[0], tuple(s[1]) if s[1] is not None else None)]
            for s in [
                l
                for l in canvas.collections
                if isinstance(l, mpl.collections.LineCollection)
            ][0].get_linestyles()
        ]
    assert styles == expected
    plt.close()


def test_display_node_labels():
    G = nx.path_graph(4)
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas, node_label={"size": 20})
    labels = [t for t in canvas.get_children() if isinstance(t, mpl.text.Text)]
    for n, l in zip(G.nodes(), labels):
        assert l.get_text() == str(n)
        assert l.get_size() == 20.0
    plt.close()


def test_display_edge_labels():
    G = nx.path_graph(4)
    canvas = plt.figure().add_subplot(111)
    # While we can pass in dicts for edge label defaults without errors,
    # this isn't helpful unless we want one label for all edges.
    nx.set_edge_attributes(G, {(u, v): {"label": u + v} for u, v in G.edges()})
    nx.display(G, canvas=canvas, edge_label={"color": "r"}, node_label=None)
    labels = [t for t in canvas.get_children() if isinstance(t, mpl.text.Text)]
    print(labels)
    for e, l in zip(G.edges(), labels):
        assert l.get_text() == str(e[0] + e[1])
        assert l.get_color() == "r"
    plt.close()


def test_display_multigraph_non_integer_keys():
    G = nx.MultiGraph()
    G.add_nodes_from(["A", "B", "C", "D"])
    G.add_edges_from(
        [
            ("A", "B", "0"),
            ("A", "B", "1"),
            ("B", "C", "-1"),
            ("B", "C", "1"),
            ("C", "D", "-1"),
            ("C", "D", "0"),
        ]
    )
    nx.set_edge_attributes(
        G, {e: f"arc3,rad={0.2 * int(e[2])}" for e in G.edges(keys=True)}, "curvature"
    )
    canvas = plt.figure().add_subplot(111)
    nx.display(G, canvas=canvas)
    rads = [
        f.get_connectionstyle().rad
        for f in canvas.get_children()
        if isinstance(f, mpl.patches.FancyArrowPatch)
    ]
    assert rads == [0.0, 0.2, -0.2, 0.2, -0.2, 0.0]
    plt.close()


def test_display_raises_for_bad_arg():
    G = nx.karate_club_graph()
    with pytest.raises(nx.NetworkXError):
        nx.display(G, bad_arg="bad_arg")
        plt.close()


def test_display_arrow_size():
    G = nx.path_graph(4, create_using=nx.DiGraph)
    nx.set_edge_attributes(
        G, {(u, v): (u + v + 2) ** 2 for u, v in G.edges()}, "arrowsize"
    )
    ax = plt.axes()
    nx.display(G, canvas=ax)
    assert [9, 25, 49] == [
        f.get_mutation_scale()
        for f in ax.get_children()
        if isinstance(f, mpl.patches.FancyArrowPatch)
    ]
    plt.close()


def test_display_mismatched_edge_position():
    """
    This test ensures that a error is raised for incomplete position data.
    """
    G = nx.path_graph(5)
    # Notice that there is no position for node 3
    nx.set_node_attributes(G, {0: (0, 0), 1: (1, 1), 2: (2, 2), 4: (4, 4)}, "pos")
    # But that's not a problem since we don't want to show node 4, right?
    nx.set_node_attributes(G, {n: n < 4 for n in G.nodes()}, "visible")
    # However, if we try to visualize every edge (including 3 -> 4)...
    # That's a problem since node 4 doesn't have a position
    with pytest.raises(nx.NetworkXError):
        nx.display(G)


# NOTE: parametrizing on marker to test both branches of internal
# to_marker_edge function
@pytest.mark.parametrize("node_shape", ("o", "s"))
def test_display_edge_margins(node_shape):
    """
    Test that there is a wider gap between the node and the start of an
    incident edge when min_source_margin is specified.

    This test checks that the use os min_{source/target}_margin edge
    attributes result is shorter (more padding) between the edges and
    source and target nodes.


    As a crude visual example, let 's' and 't' represent source and target
    nodes, respectively:

       Default:
       s-----------------------------t

       With margins:
       s   -----------------------   t

    """
    ax = plt.figure().add_subplot(111)
    G = nx.DiGraph([(0, 1)])
    nx.set_node_attributes(G, {0: (0, 0), 1: (1, 1)}, "pos")
    # Get the default patches from the regular visualization
    nx.display(G, canvas=ax, node_shape=node_shape)
    default_arrow = [
        f for f in ax.get_children() if isinstance(f, mpl.patches.FancyArrowPatch)
    ][0]
    default_extent = default_arrow.get_extents().corners()[::2, 0]
    # Now plot again with margins
    ax = plt.figure().add_subplot(111)
    nx.display(
        G,
        canvas=ax,
        edge_source_margin=100,
        edge_target_margin=100,
        node_shape=node_shape,
    )
    padded_arrow = [
        f for f in ax.get_children() if isinstance(f, mpl.patches.FancyArrowPatch)
    ][0]
    padded_extent = padded_arrow.get_extents().corners()[::2, 0]

    # With padding, the left-most extent of the edge should be further to the right
    assert padded_extent[0] > default_extent[0]
    # And the rightmost extent of the edge, further to the left
    assert padded_extent[1] < default_extent[1]
    plt.close()


@pytest.mark.parametrize("ticks", [False, True])
def test_display_hide_ticks(ticks):
    G = nx.path_graph(3)
    nx.set_node_attributes(G, {n: (n, n) for n in G.nodes()}, "pos")
    ax = plt.axes()
    nx.display(G, hide_ticks=ticks)
    for axis in [ax.xaxis, ax.yaxis]:
        assert bool(axis.get_ticklabels()) != ticks

    plt.close()


def test_display_self_loop():
    ax = plt.axes()
    G = nx.DiGraph()
    G.add_node(0)
    G.add_edge(0, 0)
    nx.set_node_attributes(G, {0: (0, 0)}, "pos")
    nx.display(G, canvas=ax)
    arrow = [
        f for f in ax.get_children() if isinstance(f, mpl.patches.FancyArrowPatch)
    ][0]
    bbox = arrow.get_extents()
    print(bbox.width)
    print(bbox.height)
    assert bbox.width > 0 and bbox.height > 0

    plt.delaxes(ax)
    plt.close()


def test_display_remove_pos_attr():
    """
    If the pos attribute isn't provided or is a function, display computes the layout
    and adds it to the graph. We need to ensure that this new attribute is removed from
    the returned graph.
    """
    G = nx.karate_club_graph()
    nx.display(G)
    assert nx.get_node_attributes(G, "display's position attribute name") == {}


@pytest.fixture
def subplots():
    fig, ax = plt.subplots()
    yield fig, ax
    plt.delaxes(ax)
    plt.close()


@pytest.mark.parametrize(
    "function",
    [
        nx.draw_circular,
        nx.draw_kamada_kawai,
        nx.draw_planar,
        nx.draw_random,
        nx.draw_spectral,
        nx.draw_spring,
        nx.draw_shell,
        nx.draw_forceatlas2,
    ],
)
def test_draw(function, subplots, tmp_path):
    if function == nx.draw_kamada_kawai:
        pytest.importorskip("scipy", reason="draw_kamada_kawai requires scipy")
    fig, _ = subplots
    options = {"node_color": "black", "node_size": 100, "width": 3}
    function(barbell, **options)
    fig.savefig(tmp_path / "test.ps")


def test_draw_shell_nlist(subplots, tmp_path):
    fig, _ = subplots
    nlist = [list(range(4)), list(range(4, 10)), list(range(10, 14))]
    nx.draw_shell(barbell, nlist=nlist)
    fig.savefig(tmp_path / "test.ps")


def test_draw_bipartite(subplots, tmp_path):
    fig, _ = subplots
    G = nx.complete_bipartite_graph(2, 5)
    nx.draw_bipartite(G)
    fig.savefig(tmp_path / "test.ps")


def test_edge_colormap():
    colors = range(barbell.number_of_edges())
    nx.draw_spring(
        barbell, edge_color=colors, width=4, edge_cmap=plt.cm.Blues, with_labels=True
    )
    # plt.show()


def test_draw_networkx_edge_labels(subplots, tmp_path):
    fig, _ = subplots
    edge = (0, 1)
    G = nx.DiGraph([edge])
    pos = {n: (n, n) for n in G}
    nx.draw(G, pos=pos)
    nx.draw_networkx_edge_labels(G, pos, edge_labels={edge: "edge"})
    fig.savefig(tmp_path / "test.ps")


def test_arrows():
    nx.draw_spring(barbell.to_directed())
    # plt.show()


@pytest.mark.parametrize(
    ("edge_color", "expected"),
    (
        (None, "black"),  # Default
        ("r", "red"),  # Non-default color string
        (["r"], "red"),  # Single non-default color in a list
        ((1.0, 1.0, 0.0), "yellow"),  # single color as rgb tuple
        ([(1.0, 1.0, 0.0)], "yellow"),  # single color as rgb tuple in list
        ((0, 1, 0, 1), "lime"),  # single color as rgba tuple
        ([(0, 1, 0, 1)], "lime"),  # single color as rgba tuple in list
        ("#0000ff", "blue"),  # single color hex code
        (["#0000ff"], "blue"),  # hex code in list
    ),
)
@pytest.mark.parametrize("edgelist", (None, [(0, 1)]))
def test_single_edge_color_undirected(edge_color, expected, edgelist):
    """Tests ways of specifying all edges have a single color for edges
    drawn with a LineCollection"""

    G = nx.path_graph(3)
    drawn_edges = nx.draw_networkx_edges(
        G, pos=nx.random_layout(G), edgelist=edgelist, edge_color=edge_color
    )
    assert mpl.colors.same_color(drawn_edges.get_color(), expected)


@pytest.mark.parametrize(
    ("edge_color", "expected"),
    (
        (None, "black"),  # Default
        ("r", "red"),  # Non-default color string
        (["r"], "red"),  # Single non-default color in a list
        ((1.0, 1.0, 0.0), "yellow"),  # single color as rgb tuple
        ([(1.0, 1.0, 0.0)], "yellow"),  # single color as rgb tuple in list
        ((0, 1, 0, 1), "lime"),  # single color as rgba tuple
        ([(0, 1, 0, 1)], "lime"),  # single color as rgba tuple in list
        ("#0000ff", "blue"),  # single color hex code
        (["#0000ff"], "blue"),  # hex code in list
    ),
)
@pytest.mark.parametrize("edgelist", (None, [(0, 1)]))
def test_single_edge_color_directed(edge_color, expected, edgelist):
    """Tests ways of specifying all edges have a single color for edges drawn
    with FancyArrowPatches"""

    G = nx.path_graph(3, create_using=nx.DiGraph)
    drawn_edges = nx.draw_networkx_edges(
        G, pos=nx.random_layout(G), edgelist=edgelist, edge_color=edge_color
    )
    for fap in drawn_edges:
        assert mpl.colors.same_color(fap.get_edgecolor(), expected)


def test_edge_color_tuple_interpretation():
    """If edge_color is a sequence with the same length as edgelist, then each
    value in edge_color is mapped onto each edge via colormap."""
    G = nx.path_graph(6, create_using=nx.DiGraph)
    pos = {n: (n, n) for n in range(len(G))}

    # num edges != 3 or 4 --> edge_color interpreted as rgb(a)
    for ec in ((0, 0, 1), (0, 0, 1, 1)):
        # More than 4 edges
        drawn_edges = nx.draw_networkx_edges(G, pos, edge_color=ec)
        for fap in drawn_edges:
            assert mpl.colors.same_color(fap.get_edgecolor(), ec)
        # Fewer than 3 edges
        drawn_edges = nx.draw_networkx_edges(
            G, pos, edgelist=[(0, 1), (1, 2)], edge_color=ec
        )
        for fap in drawn_edges:
            assert mpl.colors.same_color(fap.get_edgecolor(), ec)

    # num edges == 3, len(edge_color) == 4: interpreted as rgba
    drawn_edges = nx.draw_networkx_edges(
        G, pos, edgelist=[(0, 1), (1, 2), (2, 3)], edge_color=(0, 0, 1, 1)
    )
    for fap in drawn_edges:
        assert mpl.colors.same_color(fap.get_edgecolor(), "blue")

    # num edges == 4, len(edge_color) == 3: interpreted as rgb
    drawn_edges = nx.draw_networkx_edges(
        G, pos, edgelist=[(0, 1), (1, 2), (2, 3), (3, 4)], edge_color=(0, 0, 1)
    )
    for fap in drawn_edges:
        assert mpl.colors.same_color(fap.get_edgecolor(), "blue")

    # num edges == len(edge_color) == 3: interpreted with cmap, *not* as rgb
    drawn_edges = nx.draw_networkx_edges(
        G, pos, edgelist=[(0, 1), (1, 2), (2, 3)], edge_color=(0, 0, 1)
    )
    assert mpl.colors.same_color(
        drawn_edges[0].get_edgecolor(), drawn_edges[1].get_edgecolor()
    )
    for fap in drawn_edges:
        assert not mpl.colors.same_color(fap.get_edgecolor(), "blue")

    # num edges == len(edge_color) == 4: interpreted with cmap, *not* as rgba
    drawn_edges = nx.draw_networkx_edges(
        G, pos, edgelist=[(0, 1), (1, 2), (2, 3), (3, 4)], edge_color=(0, 0, 1, 1)
    )
    assert mpl.colors.same_color(
        drawn_edges[0].get_edgecolor(), drawn_edges[1].get_edgecolor()
    )
    assert mpl.colors.same_color(
        drawn_edges[2].get_edgecolor(), drawn_edges[3].get_edgecolor()
    )
    for fap in drawn_edges:
        assert not mpl.colors.same_color(fap.get_edgecolor(), "blue")


def test_fewer_edge_colors_than_num_edges_directed():
    """Test that the edge colors are cycled when there are fewer specified
    colors than edges."""
    G = barbell.to_directed()
    pos = nx.random_layout(barbell)
    edgecolors = ("r", "g", "b")
    drawn_edges = nx.draw_networkx_edges(G, pos, edge_color=edgecolors)
    for fap, expected in zip(drawn_edges, itertools.cycle(edgecolors)):
        assert mpl.colors.same_color(fap.get_edgecolor(), expected)


def test_more_edge_colors_than_num_edges_directed():
    """Test that extra edge colors are ignored when there are more specified
    colors than edges."""
    G = nx.path_graph(4, create_using=nx.DiGraph)  # 3 edges
    pos = nx.random_layout(barbell)
    edgecolors = ("r", "g", "b", "c")  # 4 edge colors
    drawn_edges = nx.draw_networkx_edges(G, pos, edge_color=edgecolors)
    for fap, expected in zip(drawn_edges, edgecolors[:-1]):
        assert mpl.colors.same_color(fap.get_edgecolor(), expected)


def test_edge_color_string_with_global_alpha_undirected():
    edge_collection = nx.draw_networkx_edges(
        barbell,
        pos=nx.random_layout(barbell),
        edgelist=[(0, 1), (1, 2)],
        edge_color="purple",
        alpha=0.2,
    )
    ec = edge_collection.get_color().squeeze()  # as rgba tuple
    assert len(edge_collection.get_paths()) == 2
    assert mpl.colors.same_color(ec[:-1], "purple")
    assert ec[-1] == 0.2


def test_edge_color_string_with_global_alpha_directed():
    drawn_edges = nx.draw_networkx_edges(
        barbell.to_directed(),
        pos=nx.random_layout(barbell),
        edgelist=[(0, 1), (1, 2)],
        edge_color="purple",
        alpha=0.2,
    )
    assert len(drawn_edges) == 2
    for fap in drawn_edges:
        ec = fap.get_edgecolor()  # As rgba tuple
        assert mpl.colors.same_color(ec[:-1], "purple")
        assert ec[-1] == 0.2


@pytest.mark.parametrize("graph_type", (nx.Graph, nx.DiGraph))
def test_edge_width_default_value(graph_type):
    """Test the default linewidth for edges drawn either via LineCollection or
    FancyArrowPatches."""
    G = nx.path_graph(2, create_using=graph_type)
    pos = {n: (n, n) for n in range(len(G))}
    drawn_edges = nx.draw_networkx_edges(G, pos)
    if isinstance(drawn_edges, list):  # directed case: list of FancyArrowPatch
        drawn_edges = drawn_edges[0]
    assert drawn_edges.get_linewidth() == 1


@pytest.mark.parametrize(
    ("edgewidth", "expected"),
    (
        (3, 3),  # single-value, non-default
        ([3], 3),  # Single value as a list
    ),
)
def test_edge_width_single_value_undirected(edgewidth, expected):
    G = nx.path_graph(4)
    pos = {n: (n, n) for n in range(len(G))}
    drawn_edges = nx.draw_networkx_edges(G, pos, width=edgewidth)
    assert len(drawn_edges.get_paths()) == 3
    assert drawn_edges.get_linewidth() == expected


@pytest.mark.parametrize(
    ("edgewidth", "expected"),
    (
        (3, 3),  # single-value, non-default
        ([3], 3),  # Single value as a list
    ),
)
def test_edge_width_single_value_directed(edgewidth, expected):
    G = nx.path_graph(4, create_using=nx.DiGraph)
    pos = {n: (n, n) for n in range(len(G))}
    drawn_edges = nx.draw_networkx_edges(G, pos, width=edgewidth)
    assert len(drawn_edges) == 3
    for fap in drawn_edges:
        assert fap.get_linewidth() == expected


@pytest.mark.parametrize(
    "edgelist",
    (
        [(0, 1), (1, 2), (2, 3)],  # one width specification per edge
        None,  #  fewer widths than edges - widths cycle
        [(0, 1), (1, 2)],  # More widths than edges - unused widths ignored
    ),
)
def test_edge_width_sequence(edgelist):
    G = barbell.to_directed()
    pos = nx.random_layout(G)
    widths = (0.5, 2.0, 12.0)
    drawn_edges = nx.draw_networkx_edges(G, pos, edgelist=edgelist, width=widths)
    for fap, expected_width in zip(drawn_edges, itertools.cycle(widths)):
        assert fap.get_linewidth() == expected_width


def test_edge_color_with_edge_vmin_vmax():
    """Test that edge_vmin and edge_vmax properly set the dynamic range of the
    color map when num edges == len(edge_colors)."""
    G = nx.path_graph(3, create_using=nx.DiGraph)
    pos = nx.random_layout(G)
    # Extract colors from the original (unscaled) colormap
    drawn_edges = nx.draw_networkx_edges(G, pos, edge_color=[0, 1.0])
    orig_colors = [e.get_edgecolor() for e in drawn_edges]
    # Colors from scaled colormap
    drawn_edges = nx.draw_networkx_edges(
        G, pos, edge_color=[0.2, 0.8], edge_vmin=0.2, edge_vmax=0.8
    )
    scaled_colors = [e.get_edgecolor() for e in drawn_edges]
    assert mpl.colors.same_color(orig_colors, scaled_colors)


def test_directed_edges_linestyle_default():
    """Test default linestyle for edges drawn with FancyArrowPatches."""
    G = nx.path_graph(4, create_using=nx.DiGraph)  # Graph with 3 edges
    pos = {n: (n, n) for n in range(len(G))}

    # edge with default style
    drawn_edges = nx.draw_networkx_edges(G, pos)
    assert len(drawn_edges) == 3
    for fap in drawn_edges:
        assert fap.get_linestyle() == "solid"


@pytest.mark.parametrize(
    "style",
    (
        "dashed",  # edge with string style
        "--",  # edge with simplified string style
        (1, (1, 1)),  # edge with (offset, onoffseq) style
    ),
)
def test_directed_edges_linestyle_single_value(style):
    """Tests support for specifying linestyles with a single value to be applied to
    all edges in ``draw_networkx_edges`` for FancyArrowPatch outputs
    (e.g. directed edges)."""

    G = nx.path_graph(4, create_using=nx.DiGraph)  # Graph with 3 edges
    pos = {n: (n, n) for n in range(len(G))}

    drawn_edges = nx.draw_networkx_edges(G, pos, style=style)
    assert len(drawn_edges) == 3
    for fap in drawn_edges:
        assert fap.get_linestyle() == style


@pytest.mark.parametrize(
    "style_seq",
    (
        ["dashed"],  # edge with string style in list
        ["--"],  # edge with simplified string style in list
        [(1, (1, 1))],  # edge with (offset, onoffseq) style in list
        ["--", "-", ":"],  # edges with styles for each edge
        ["--", "-"],  # edges with fewer styles than edges (styles cycle)
        ["--", "-", ":", "-."],  # edges with more styles than edges (extra unused)
    ),
)
def test_directed_edges_linestyle_sequence(style_seq):
    """Tests support for specifying linestyles with sequences in
    ``draw_networkx_edges`` for FancyArrowPatch outputs (e.g. directed edges)."""

    G = nx.path_graph(4, create_using=nx.DiGraph)  # Graph with 3 edges
    pos = {n: (n, n) for n in range(len(G))}

    drawn_edges = nx.draw_networkx_edges(G, pos, style=style_seq)
    assert len(drawn_edges) == 3
    for fap, style in zip(drawn_edges, itertools.cycle(style_seq)):
        assert fap.get_linestyle() == style


def test_return_types():
    from matplotlib.collections import LineCollection, PathCollection
    from matplotlib.patches import FancyArrowPatch

    G = nx.frucht_graph(create_using=nx.Graph)
    dG = nx.frucht_graph(create_using=nx.DiGraph)
    pos = nx.spring_layout(G, seed=42)
    dpos = nx.spring_layout(dG, seed=42)
    # nodes
    nodes = nx.draw_networkx_nodes(G, pos)
    assert isinstance(nodes, PathCollection)
    # edges
    edges = nx.draw_networkx_edges(dG, dpos, arrows=True)
    assert isinstance(edges, list)
    if len(edges) > 0:
        assert isinstance(edges[0], FancyArrowPatch)
    edges = nx.draw_networkx_edges(dG, dpos, arrows=False)
    assert isinstance(edges, LineCollection)
    edges = nx.draw_networkx_edges(G, dpos, arrows=None)
    assert isinstance(edges, LineCollection)
    edges = nx.draw_networkx_edges(dG, pos, arrows=None)
    assert isinstance(edges, list)
    if len(edges) > 0:
        assert isinstance(edges[0], FancyArrowPatch)


def test_labels_and_colors():
    G = nx.cubical_graph()
    pos = nx.spring_layout(G, seed=42)  # positions for all nodes
    # nodes
    nx.draw_networkx_nodes(
        G, pos, nodelist=[0, 1, 2, 3], node_color="r", node_size=500, alpha=0.75
    )
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=[4, 5, 6, 7],
        node_color="b",
        node_size=500,
        alpha=[0.25, 0.5, 0.75, 1.0],
    )
    # edges
    nx.draw_networkx_edges(G, pos, width=1.0, alpha=0.5)
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=[(0, 1), (1, 2), (2, 3), (3, 0)],
        width=8,
        alpha=0.5,
        edge_color="r",
    )
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=[(4, 5), (5, 6), (6, 7), (7, 4)],
        width=8,
        alpha=0.5,
        edge_color="b",
    )
    nx.draw_networkx_edges(
        G,
        pos,
        edgelist=[(4, 5), (5, 6), (6, 7), (7, 4)],
        arrows=True,
        min_source_margin=0.5,
        min_target_margin=0.75,
        width=8,
        edge_color="b",
    )
    # some math labels
    labels = {}
    labels[0] = r"$a$"
    labels[1] = r"$b$"
    labels[2] = r"$c$"
    labels[3] = r"$d$"
    labels[4] = r"$\alpha$"
    labels[5] = r"$\beta$"
    labels[6] = r"$\gamma$"
    labels[7] = r"$\delta$"
    colors = {n: "k" if n % 2 == 0 else "r" for n in range(8)}
    nx.draw_networkx_labels(G, pos, labels, font_size=16)
    nx.draw_networkx_labels(G, pos, labels, font_size=16, font_color=colors)
    nx.draw_networkx_edge_labels(G, pos, edge_labels=None, rotate=False)
    nx.draw_networkx_edge_labels(G, pos, edge_labels={(4, 5): "4-5"})
    # plt.show()


def test_axes(subplots):
    fig, ax = subplots
    nx.draw(barbell, ax=ax)
    nx.draw_networkx_edge_labels(barbell, nx.circular_layout(barbell), ax=ax)


def test_empty_graph():
    G = nx.Graph()
    nx.draw(G)


def test_draw_empty_nodes_return_values():
    # See Issue #3833
    import matplotlib.collections  # call as mpl.collections

    G = nx.Graph([(1, 2), (2, 3)])
    DG = nx.DiGraph([(1, 2), (2, 3)])
    pos = nx.circular_layout(G)
    assert isinstance(
        nx.draw_networkx_nodes(G, pos, nodelist=[]), mpl.collections.PathCollection
    )
    assert isinstance(
        nx.draw_networkx_nodes(DG, pos, nodelist=[]), mpl.collections.PathCollection
    )

    # drawing empty edges used to return an empty LineCollection or empty list.
    # Now it is always an empty list (because edges are now lists of FancyArrows)
    assert nx.draw_networkx_edges(G, pos, edgelist=[], arrows=True) == []
    assert nx.draw_networkx_edges(G, pos, edgelist=[], arrows=False) == []
    assert nx.draw_networkx_edges(DG, pos, edgelist=[], arrows=False) == []
    assert nx.draw_networkx_edges(DG, pos, edgelist=[], arrows=True) == []


def test_multigraph_edgelist_tuples():
    # See Issue #3295
    G = nx.path_graph(3, create_using=nx.MultiDiGraph)
    nx.draw_networkx(G, edgelist=[(0, 1, 0)])
    nx.draw_networkx(G, edgelist=[(0, 1, 0)], node_size=[10, 20, 0])


def test_alpha_iter():
    pos = nx.random_layout(barbell)
    fig = plt.figure()
    # with fewer alpha elements than nodes
    fig.add_subplot(131)  # Each test in a new axis object
    nx.draw_networkx_nodes(barbell, pos, alpha=[0.1, 0.2])
    # with equal alpha elements and nodes
    num_nodes = len(barbell.nodes)
    alpha = [x / num_nodes for x in range(num_nodes)]
    colors = range(num_nodes)
    fig.add_subplot(132)
    nx.draw_networkx_nodes(barbell, pos, node_color=colors, alpha=alpha)
    # with more alpha elements than nodes
    alpha.append(1)
    fig.add_subplot(133)
    nx.draw_networkx_nodes(barbell, pos, alpha=alpha)


def test_multiple_node_shapes(subplots):
    fig, ax = subplots
    G = nx.path_graph(4)
    nx.draw(G, node_shape=["o", "h", "s", "^"], ax=ax)
    scatters = [
        s for s in ax.get_children() if isinstance(s, mpl.collections.PathCollection)
    ]
    assert len(scatters) == 4


def test_individualized_font_attributes(subplots):
    G = nx.karate_club_graph()
    fig, ax = subplots
    nx.draw(
        G,
        ax=ax,
        font_color={n: "k" if n % 2 else "r" for n in G.nodes()},
        font_size={n: int(n / (34 / 15) + 5) for n in G.nodes()},
    )
    for n, t in zip(
        G.nodes(),
        [
            t
            for t in ax.get_children()
            if isinstance(t, mpl.text.Text) and len(t.get_text()) > 0
        ],
    ):
        expected = "black" if n % 2 else "red"

        assert mpl.colors.same_color(t.get_color(), expected)
        assert int(n / (34 / 15) + 5) == t.get_size()


def test_individualized_edge_attributes(subplots):
    G = nx.karate_club_graph()
    fig, ax = subplots
    arrowstyles = ["-|>" if (u + v) % 2 == 0 else "-[" for u, v in G.edges()]
    arrowsizes = [10 * (u % 2 + v % 2) + 10 for u, v in G.edges()]
    nx.draw(G, ax=ax, arrows=True, arrowstyle=arrowstyles, arrowsize=arrowsizes)
    arrows = [
        f for f in ax.get_children() if isinstance(f, mpl.patches.FancyArrowPatch)
    ]
    for e, a in zip(G.edges(), arrows):
        assert a.get_mutation_scale() == 10 * (e[0] % 2 + e[1] % 2) + 10
        expected = (
            mpl.patches.ArrowStyle.BracketB
            if sum(e) % 2
            else mpl.patches.ArrowStyle.CurveFilledB
        )
        assert isinstance(a.get_arrowstyle(), expected)


def test_error_invalid_kwds():
    with pytest.raises(ValueError, match="Received invalid argument"):
        nx.draw(barbell, foo="bar")


def test_draw_networkx_arrowsize_incorrect_size():
    G = nx.DiGraph([(0, 1), (0, 2), (0, 3), (1, 3)])
    arrowsize = [1, 2, 3]
    with pytest.raises(
        ValueError, match="arrowsize should have the same length as edgelist"
    ):
        nx.draw(G, arrowsize=arrowsize)


@pytest.mark.parametrize("arrowsize", (30, [10, 20, 30]))
def test_draw_edges_arrowsize(arrowsize):
    G = nx.DiGraph([(0, 1), (0, 2), (1, 2)])
    pos = {0: (0, 0), 1: (0, 1), 2: (1, 0)}
    edges = nx.draw_networkx_edges(G, pos=pos, arrowsize=arrowsize)

    arrowsize = itertools.repeat(arrowsize) if isinstance(arrowsize, int) else arrowsize

    for fap, expected in zip(edges, arrowsize):
        assert isinstance(fap, mpl.patches.FancyArrowPatch)
        assert fap.get_mutation_scale() == expected


@pytest.mark.parametrize("arrowstyle", ("-|>", ["-|>", "-[", "<|-|>"]))
def test_draw_edges_arrowstyle(arrowstyle):
    G = nx.DiGraph([(0, 1), (0, 2), (1, 2)])
    pos = {0: (0, 0), 1: (0, 1), 2: (1, 0)}
    edges = nx.draw_networkx_edges(G, pos=pos, arrowstyle=arrowstyle)

    arrowstyle = (
        itertools.repeat(arrowstyle) if isinstance(arrowstyle, str) else arrowstyle
    )

    arrow_objects = {
        "-|>": mpl.patches.ArrowStyle.CurveFilledB,
        "-[": mpl.patches.ArrowStyle.BracketB,
        "<|-|>": mpl.patches.ArrowStyle.CurveFilledAB,
    }

    for fap, expected in zip(edges, arrowstyle):
        assert isinstance(fap, mpl.patches.FancyArrowPatch)
        assert isinstance(fap.get_arrowstyle(), arrow_objects[expected])


def test_np_edgelist():
    # see issue #4129
    nx.draw_networkx(barbell, edgelist=np.array([(0, 2), (0, 3)]))


def test_draw_nodes_missing_node_from_position():
    G = nx.path_graph(3)
    pos = {0: (0, 0), 1: (1, 1)}  # No position for node 2
    with pytest.raises(nx.NetworkXError, match="has no position"):
        nx.draw_networkx_nodes(G, pos)


def test_draw_networkx_nodes_node_shape_list_with_scalar_color(subplots):
    """Ensure draw_networkx_nodes works when node_shape is a Python list.

    This covers the case where node_shape is a sequence (list) and node_color
    is a single scalar color, which should be supported.
    """
    fig, ax = subplots

    G = nx.empty_graph(5)
    pos = {i: (i, i) for i in G}

    shapes = ["o", "^", "o", "^", "o"]

    nodes = nx.draw_networkx_nodes(
        G,
        pos,
        node_color="red",  # scalar color (supported)
        node_shape=shapes,  # list of shapes – this used to be buggy
        ax=ax,
    )
    # Should get a PathCollection with an element in it (same as with numpy arrays)
    assert len(nodes.get_offsets()) > 0
    # NOTE: When node_shape is a sequence, draw_networkx_nodes internally calls
    # ax.scatter multiple times and returns only the last PathCollection.
    # Therefore, we do NOT assert a value for len(nodes.get_offsets()) here.


# NOTE: parametrizing on marker to test both branches of internal
# nx.draw_networkx_edges.to_marker_edge function
@pytest.mark.parametrize("node_shape", ("o", "s"))
def test_draw_edges_min_source_target_margins(node_shape, subplots):
    """Test that there is a wider gap between the node and the start of an
    incident edge when min_source_margin is specified.

    This test checks that the use of min_{source/target}_margin kwargs result
    in shorter (more padding) between the edges and source and target nodes.
    As a crude visual example, let 's' and 't' represent source and target
    nodes, respectively:

       Default:
       s-----------------------------t

       With margins:
       s   -----------------------   t

    """
    # Create a single axis object to get consistent pixel coords across
    # multiple draws
    fig, ax = subplots
    G = nx.DiGraph([(0, 1)])
    pos = {0: (0, 0), 1: (1, 0)}  # horizontal layout
    # Get leftmost and rightmost points of the FancyArrowPatch object
    # representing the edge between nodes 0 and 1 (in pixel coordinates)
    default_patch = nx.draw_networkx_edges(G, pos, ax=ax, node_shape=node_shape)[0]
    default_extent = default_patch.get_extents().corners()[::2, 0]
    # Now, do the same but with "padding" for the source and target via the
    # min_{source/target}_margin kwargs
    padded_patch = nx.draw_networkx_edges(
        G,
        pos,
        ax=ax,
        node_shape=node_shape,
        min_source_margin=100,
        min_target_margin=100,
    )[0]
    padded_extent = padded_patch.get_extents().corners()[::2, 0]

    # With padding, the left-most extent of the edge should be further to the
    # right
    assert padded_extent[0] > default_extent[0]
    # And the rightmost extent of the edge, further to the left
    assert padded_extent[1] < default_extent[1]


# NOTE: parametrizing on marker to test both branches of internal
# nx.draw_networkx_edges.to_marker_edge function
@pytest.mark.parametrize("node_shape", ("o", "s"))
def test_draw_edges_min_source_target_margins_individual(node_shape, subplots):
    """Test that there is a wider gap between the node and the start of an
    incident edge when min_source_margin is specified.

    This test checks that the use of min_{source/target}_margin kwargs result
    in shorter (more padding) between the edges and source and target nodes.
    As a crude visual example, let 's' and 't' represent source and target
    nodes, respectively:

       Default:
       s-----------------------------t

       With margins:
       s   -----------------------   t

    """
    # Create a single axis object to get consistent pixel coords across
    # multiple draws
    fig, ax = subplots
    G = nx.DiGraph([(0, 1), (1, 2)])
    pos = {0: (0, 0), 1: (1, 0), 2: (2, 0)}  # horizontal layout
    # Get leftmost and rightmost points of the FancyArrowPatch object
    # representing the edge between nodes 0 and 1 (in pixel coordinates)
    default_patch = nx.draw_networkx_edges(G, pos, ax=ax, node_shape=node_shape)
    default_extent = [d.get_extents().corners()[::2, 0] for d in default_patch]
    # Now, do the same but with "padding" for the source and target via the
    # min_{source/target}_margin kwargs
    padded_patch = nx.draw_networkx_edges(
        G,
        pos,
        ax=ax,
        node_shape=node_shape,
        min_source_margin=[98, 102],
        min_target_margin=[98, 102],
    )
    padded_extent = [p.get_extents().corners()[::2, 0] for p in padded_patch]
    for d, p in zip(default_extent, padded_extent):
        # With padding, the left-most extent of the edge should be further to the
        # right
        assert p[0] > d[0]
        # And the rightmost extent of the edge, further to the left
        assert p[1] < d[1]


def test_nonzero_selfloop_with_single_node(subplots):
    """Ensure that selfloop extent is non-zero when there is only one node."""
    # Create explicit axis object for test
    fig, ax = subplots
    # Graph with single node + self loop
    G = nx.DiGraph()
    G.add_node(0)
    G.add_edge(0, 0)
    # Draw
    patch = nx.draw_networkx_edges(G, {0: (0, 0)})[0]
    # The resulting patch must have non-zero extent
    bbox = patch.get_extents()
    assert bbox.width > 0 and bbox.height > 0


def test_nonzero_selfloop_with_single_edge_in_edgelist(subplots):
    """Ensure that selfloop extent is non-zero when only a single edge is
    specified in the edgelist.
    """
    # Create explicit axis object for test
    fig, ax = subplots
    # Graph with selfloop
    G = nx.path_graph(2, create_using=nx.DiGraph)
    G.add_edge(1, 1)
    pos = {n: (n, n) for n in G.nodes}
    # Draw only the selfloop edge via the `edgelist` kwarg
    patch = nx.draw_networkx_edges(G, pos, edgelist=[(1, 1)])[0]
    # The resulting patch must have non-zero extent
    bbox = patch.get_extents()
    assert bbox.width > 0 and bbox.height > 0


def test_apply_alpha():
    """Test apply_alpha when there is a mismatch between the number of
    supplied colors and elements.
    """
    nodelist = [0, 1, 2]
    colorlist = ["r", "g", "b"]
    alpha = 0.5
    rgba_colors = nx.drawing.nx_pylab.apply_alpha(colorlist, alpha, nodelist)
    assert all(rgba_colors[:, -1] == alpha)


def test_draw_edges_toggling_with_arrows_kwarg():
    """
    The `arrows` keyword argument is used as a 3-way switch to select which
    type of object to use for drawing edges:
      - ``arrows=None`` -> default (FancyArrowPatches for directed, else LineCollection)
      - ``arrows=True`` -> FancyArrowPatches
      - ``arrows=False`` -> LineCollection
    """
    import matplotlib.collections
    import matplotlib.patches

    UG = nx.path_graph(3)
    DG = nx.path_graph(3, create_using=nx.DiGraph)
    pos = {n: (n, n) for n in UG}

    # Use FancyArrowPatches when arrows=True, regardless of graph type
    for G in (UG, DG):
        edges = nx.draw_networkx_edges(G, pos, arrows=True)
        assert len(edges) == len(G.edges)
        assert isinstance(edges[0], mpl.patches.FancyArrowPatch)

    # Use LineCollection when arrows=False, regardless of graph type
    for G in (UG, DG):
        edges = nx.draw_networkx_edges(G, pos, arrows=False)
        assert isinstance(edges, mpl.collections.LineCollection)

    # Default behavior when arrows=None: FAPs for directed, LC's for undirected
    edges = nx.draw_networkx_edges(UG, pos)
    assert isinstance(edges, mpl.collections.LineCollection)
    edges = nx.draw_networkx_edges(DG, pos)
    assert len(edges) == len(G.edges)
    assert isinstance(edges[0], mpl.patches.FancyArrowPatch)


@pytest.mark.parametrize("drawing_func", (nx.draw, nx.draw_networkx))
def test_draw_networkx_arrows_default_undirected(drawing_func, subplots):
    import matplotlib.collections

    G = nx.path_graph(3)
    fig, ax = subplots
    drawing_func(G, ax=ax)
    assert any(isinstance(c, mpl.collections.LineCollection) for c in ax.collections)
    assert not ax.patches


@pytest.mark.parametrize("drawing_func", (nx.draw, nx.draw_networkx))
def test_draw_networkx_arrows_default_directed(drawing_func, subplots):
    import matplotlib.collections

    G = nx.path_graph(3, create_using=nx.DiGraph)
    fig, ax = subplots
    drawing_func(G, ax=ax)
    assert not any(
        isinstance(c, mpl.collections.LineCollection) for c in ax.collections
    )
    assert ax.patches


def test_edgelist_kwarg_not_ignored(subplots):
    # See gh-4994
    G = nx.path_graph(3)
    G.add_edge(0, 0)
    fig, ax = subplots
    nx.draw(G, edgelist=[(0, 1), (1, 2)], ax=ax)  # Exclude self-loop from edgelist
    assert not ax.patches


@pytest.mark.parametrize(
    ("G", "expected_n_edges"),
    ([nx.DiGraph(), 2], [nx.MultiGraph(), 4], [nx.MultiDiGraph(), 4]),
)
def test_draw_networkx_edges_multiedge_connectionstyle(G, expected_n_edges):
    """Draws edges correctly for 3 types of graphs and checks for valid length"""
    for i, (u, v) in enumerate([(0, 1), (0, 1), (0, 1), (0, 2)]):
        G.add_edge(u, v, weight=round(i / 3, 2))
    pos = {n: (n, n) for n in G}
    # Raises on insufficient connectionstyle length
    for conn_style in [
        "arc3,rad=0.1",
        ["arc3,rad=0.1", "arc3,rad=0.1"],
        ["arc3,rad=0.1", "arc3,rad=0.1", "arc3,rad=0.2"],
    ]:
        nx.draw_networkx_edges(G, pos, connectionstyle=conn_style)
        arrows = nx.draw_networkx_edges(G, pos, connectionstyle=conn_style)
        assert len(arrows) == expected_n_edges


@pytest.mark.parametrize(
    ("G", "expected_n_edges"),
    ([nx.DiGraph(), 2], [nx.MultiGraph(), 4], [nx.MultiDiGraph(), 4]),
)
def test_draw_networkx_edge_labels_multiedge_connectionstyle(G, expected_n_edges):
    """Draws labels correctly for 3 types of graphs and checks for valid length and class names"""
    for i, (u, v) in enumerate([(0, 1), (0, 1), (0, 1), (0, 2)]):
        G.add_edge(u, v, weight=round(i / 3, 2))
    pos = {n: (n, n) for n in G}
    # Raises on insufficient connectionstyle length
    arrows = nx.draw_networkx_edges(
        G, pos, connectionstyle=["arc3,rad=0.1", "arc3,rad=0.1", "arc3,rad=0.1"]
    )
    for conn_style in [
        "arc3,rad=0.1",
        ["arc3,rad=0.1", "arc3,rad=0.2"],
        ["arc3,rad=0.1", "arc3,rad=0.1", "arc3,rad=0.1"],
    ]:
        text_items = nx.draw_networkx_edge_labels(G, pos, connectionstyle=conn_style)
        assert len(text_items) == expected_n_edges
        for ti in text_items.values():
            assert ti.__class__.__name__ == "CurvedArrowText"


def test_draw_networkx_edge_label_multiedge():
    G = nx.MultiGraph()
    G.add_edge(0, 1, weight=10)
    G.add_edge(0, 1, weight=20)
    edge_labels = nx.get_edge_attributes(G, "weight")  # Includes edge keys
    pos = {n: (n, n) for n in G}
    text_items = nx.draw_networkx_edge_labels(
        G,
        pos,
        edge_labels=edge_labels,
        connectionstyle=["arc3,rad=0.1", "arc3,rad=0.2"],
    )
    assert len(text_items) == 2


def test_draw_networkx_edge_label_empty_dict():
    """Regression test for draw_networkx_edge_labels with empty dict. See
    gh-5372."""
    G = nx.path_graph(3)
    pos = {n: (n, n) for n in G.nodes}
    assert nx.draw_networkx_edge_labels(G, pos, edge_labels={}) == {}


def test_draw_networkx_edges_undirected_selfloop_colors(subplots):
    """When an edgelist is supplied along with a sequence of colors, check that
    the self-loops have the correct colors."""
    fig, ax = subplots
    # Edge list and corresponding colors
    edgelist = [(1, 3), (1, 2), (2, 3), (1, 1), (3, 3), (2, 2)]
    edge_colors = ["pink", "cyan", "black", "red", "blue", "green"]

    G = nx.Graph(edgelist)
    pos = {n: (n, n) for n in G.nodes}
    nx.draw_networkx_edges(G, pos, ax=ax, edgelist=edgelist, edge_color=edge_colors)

    # Verify that there are three fancy arrow patches (1 per self loop)
    assert len(ax.patches) == 3

    # These are points that should be contained in the self loops. For example,
    # sl_points[0] will be (1, 1.1), which is inside the "path" of the first
    # self-loop but outside the others
    sl_points = np.array(edgelist[-3:]) + np.array([0, 0.1])

    # Check that the mapping between self-loop locations and their colors is
    # correct
    for fap, clr, slp in zip(ax.patches, edge_colors[-3:], sl_points):
        assert fap.get_path().contains_point(slp)
        assert mpl.colors.same_color(fap.get_edgecolor(), clr)


@pytest.mark.parametrize(
    "fap_only_kwarg",  # Non-default values for kwargs that only apply to FAPs
    (
        {"arrowstyle": "-"},
        {"arrowsize": 20},
        {"connectionstyle": "arc3,rad=0.2"},
        {"min_source_margin": 10},
        {"min_target_margin": 10},
    ),
)
def test_user_warnings_for_unused_edge_drawing_kwargs(fap_only_kwarg, subplots):
    """Users should get a warning when they specify a non-default value for
    one of the kwargs that applies only to edges drawn with FancyArrowPatches,
    but FancyArrowPatches aren't being used under the hood."""
    G = nx.path_graph(3)
    pos = {n: (n, n) for n in G}
    fig, ax = subplots
    # By default, an undirected graph will use LineCollection to represent
    # the edges
    kwarg_name = list(fap_only_kwarg.keys())[0]
    with pytest.warns(
        UserWarning, match=f"\n\nThe {kwarg_name} keyword argument is not applicable"
    ):
        nx.draw_networkx_edges(G, pos, ax=ax, **fap_only_kwarg)
    # FancyArrowPatches are always used when `arrows=True` is specified.
    # Check that warnings are *not* raised in this case
    with warnings.catch_warnings():
        # Escalate warnings -> errors so tests fail if warnings are raised
        warnings.simplefilter("error")
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        nx.draw_networkx_edges(G, pos, ax=ax, arrows=True, **fap_only_kwarg)


@pytest.mark.parametrize("draw_fn", (nx.draw, nx.draw_circular))
def test_no_warning_on_default_draw_arrowstyle(draw_fn, subplots):
    # See gh-7284
    fig, ax = subplots
    G = nx.cycle_graph(5)
    with warnings.catch_warnings(record=True) as w:
        draw_fn(G, ax=ax)
    assert len(w) == 0


@pytest.mark.parametrize("hide_ticks", [False, True])
@pytest.mark.parametrize(
    "method",
    [
        nx.draw_networkx,
        nx.draw_networkx_edge_labels,
        nx.draw_networkx_edges,
        nx.draw_networkx_labels,
        nx.draw_networkx_nodes,
    ],
)
def test_hide_ticks(method, hide_ticks, subplots):
    G = nx.path_graph(3)
    pos = {n: (n, n) for n in G.nodes}
    _, ax = subplots
    method(G, pos=pos, ax=ax, hide_ticks=hide_ticks)
    for axis in [ax.xaxis, ax.yaxis]:
        assert bool(axis.get_ticklabels()) != hide_ticks


@pytest.mark.parametrize(
    "style", ["angle", "angle3", "arc", "arc3,rad=0.0", "bar,fraction=0.1"]
)
def test_edge_label_all_connectionstyles(subplots, style):
    """
    Check that FancyArrowPatches with all `connectionstyle`s are supported
    in edge label rendering. See gh-7735 and gh-8106.
    """
    fig, ax = subplots
    edge = (0, 1)
    G = nx.DiGraph([edge])
    pos = {n: (n, 0) for n in G}

    name = style.split(",")[0]
    labels = nx.draw_networkx_edge_labels(
        G, pos, edge_labels={edge: "edge"}, connectionstyle=style
    )

    hmid = (pos[0][0] + pos[1][0]) / 2
    vmid = (pos[0][1] + pos[1][1]) / 2
    if name in {"arc", "arc3"}:  # The label should be at roughly the midpoint.
        assert labels[edge].x, labels[edge].y == pytest.approx((hmid, vmid))
    elif name == "bar":  # The label should be below the vertical midpoint.
        assert labels[edge].y < vmid


@pytest.mark.parametrize("label_pos", [-0.1, 1.1])
def test_edge_label_label_pos(subplots, label_pos):
    """
    Check that label positions can be extrapolated outside [0, 1].
    """
    fig, ax = subplots
    edge = (0, 1)
    G = nx.DiGraph([edge])
    pos = {n: (n, n) for n in G}
    lbl = nx.draw_networkx_edge_labels(
        G, pos, edge_labels={edge: "edge"}, label_pos=label_pos, connectionstyle="angle"
    )

    assert lbl[edge].x, lbl[edge].y == pytest.approx((label_pos, label_pos))



import pytest

import networkx as nx


def test_degree_sequences():
    seq = nx.utils.powerlaw_sequence(10, seed=1)
    seq = nx.utils.powerlaw_sequence(10)
    assert len(seq) == 10


@pytest.mark.parametrize(
    ("deg_seq", "valid", "reason"),
    [
        ([], False, "must have one more node"),
        ([0], True, ""),
        ([2], False, "must have one more node"),
        ([2, 0], False, "must have strictly positive"),
        ([3, 1, 1, 1], True, ""),
    ],
)
def test_valid_degree_sequence(deg_seq, valid, reason):
    v, r = nx.utils.is_valid_tree_degree_sequence(deg_seq)
    assert v == valid
    assert reason in r


def test_zipf_rv():
    r = nx.utils.zipf_rv(2.3, xmin=2, seed=1)
    r = nx.utils.zipf_rv(2.3, 2, 1)
    r = nx.utils.zipf_rv(2.3)
    assert type(r), int
    pytest.raises(ValueError, nx.utils.zipf_rv, 0.5)
    pytest.raises(ValueError, nx.utils.zipf_rv, 2, xmin=0)


def test_random_weighted_sample():
    mapping = {"a": 10, "b": 20}
    s = nx.utils.random_weighted_sample(mapping, 2, seed=1)
    s = nx.utils.random_weighted_sample(mapping, 2)
    assert sorted(s) == sorted(mapping.keys())
    pytest.raises(ValueError, nx.utils.random_weighted_sample, mapping, 3)


def test_random_weighted_choice():
    mapping = {"a": 10, "b": 0}
    c = nx.utils.weighted_choice(mapping, seed=1)
    c = nx.utils.weighted_choice(mapping)
    assert c == "a"


def test_random_sequence_low_precision():
    assert nx.utils.cumulative_distribution([0.1] * 100)[-1] == 1.0



import itertools

import pytest

import networkx as nx
from networkx.algorithms import bipartite

pytest.importorskip("scipy")


class TestBipartiteLinkAnalysis:
    @classmethod
    def setup_class(cls):
        cls.davis_southern_women_graph = nx.davis_southern_women_graph()
        cls.women_bipartite_set = {
            node
            for node, bipartite in cls.davis_southern_women_graph.nodes(
                data="bipartite"
            )
            if bipartite == 0
        }
        cls.gnmk_random_graph = nx.bipartite.generators.gnmk_random_graph(
            5 * 10**2, 10**2, 5 * 10**2, seed=27
        )
        cls.gnmk_random_graph_top_nodes = {
            node
            for node, bipartite in cls.gnmk_random_graph.nodes(data="bipartite")
            if bipartite == 0
        }

    def test_collaborative_filtering_birank(self):
        elist = [
            ("u1", "p1", 5),
            ("u2", "p1", 5),
            ("u2", "p2", 4),
            ("u3", "p1", 3),
            ("u3", "p3", 2),
        ]
        item_recommendation_graph = nx.DiGraph()
        item_recommendation_graph.add_weighted_edges_from(elist, weight="rating")
        product_nodes = ("p1", "p2", "p3")
        u1_query = {
            product: rating
            for _, product, rating in item_recommendation_graph.edges(
                nbunch="u1", data="rating"
            )
        }
        u1_birank_results = bipartite.birank(
            item_recommendation_graph,
            product_nodes,
            alpha=0.8,
            beta=1.0,
            top_personalization=u1_query,
            weight="rating",
        )

        assert u1_birank_results["p2"] > u1_birank_results["p3"]

        u1_birank_results_unweighted = bipartite.birank(
            item_recommendation_graph,
            product_nodes,
            alpha=0.8,
            beta=1.0,
            top_personalization=u1_query,
            weight=None,
        )

        assert u1_birank_results_unweighted["p2"] == pytest.approx(
            u1_birank_results_unweighted["p3"], rel=2e-6
        )

    def test_davis_birank(self):
        scores = bipartite.birank(
            self.davis_southern_women_graph, self.women_bipartite_set
        )
        answer = {
            "Laura Mandeville": 0.07,
            "Olivia Carleton": 0.04,
            "Frances Anderson": 0.05,
            "Pearl Oglethorpe": 0.04,
            "Katherina Rogers": 0.06,
            "Flora Price": 0.04,
            "Dorothy Murchison": 0.04,
            "Helen Lloyd": 0.06,
            "Theresa Anderson": 0.07,
            "Eleanor Nye": 0.05,
            "Evelyn Jefferson": 0.07,
            "Sylvia Avondale": 0.07,
            "Charlotte McDowd": 0.05,
            "Verne Sanderson": 0.05,
            "Myra Liddel": 0.05,
            "Brenda Rogers": 0.07,
            "Ruth DeSand": 0.05,
            "Nora Fayette": 0.07,
            "E8": 0.11,
            "E7": 0.09,
            "E10": 0.07,
            "E9": 0.1,
            "E13": 0.05,
            "E3": 0.07,
            "E12": 0.07,
            "E11": 0.06,
            "E2": 0.05,
            "E5": 0.08,
            "E6": 0.08,
            "E14": 0.05,
            "E4": 0.06,
            "E1": 0.05,
        }

        for node, value in answer.items():
            assert scores[node] == pytest.approx(value, abs=1e-2)

    def test_davis_birank_with_personalization(self):
        women_personalization = {"Laura Mandeville": 1}
        scores = bipartite.birank(
            self.davis_southern_women_graph,
            self.women_bipartite_set,
            top_personalization=women_personalization,
        )
        answer = {
            "Laura Mandeville": 0.29,
            "Olivia Carleton": 0.02,
            "Frances Anderson": 0.06,
            "Pearl Oglethorpe": 0.04,
            "Katherina Rogers": 0.04,
            "Flora Price": 0.02,
            "Dorothy Murchison": 0.03,
            "Helen Lloyd": 0.04,
            "Theresa Anderson": 0.08,
            "Eleanor Nye": 0.05,
            "Evelyn Jefferson": 0.09,
            "Sylvia Avondale": 0.05,
            "Charlotte McDowd": 0.06,
            "Verne Sanderson": 0.04,
            "Myra Liddel": 0.03,
            "Brenda Rogers": 0.08,
            "Ruth DeSand": 0.05,
            "Nora Fayette": 0.05,
            "E8": 0.11,
            "E7": 0.1,
            "E10": 0.04,
            "E9": 0.07,
            "E13": 0.03,
            "E3": 0.11,
            "E12": 0.04,
            "E11": 0.03,
            "E2": 0.1,
            "E5": 0.11,
            "E6": 0.1,
            "E14": 0.03,
            "E4": 0.06,
            "E1": 0.1,
        }

        for node, value in answer.items():
            assert scores[node] == pytest.approx(value, abs=1e-2)

    def test_birank_empty_bipartite_set(self):
        G = nx.Graph()
        all_nodes = [1, 2, 3]
        G.add_nodes_from(all_nodes)

        # Test with empty bipartite set
        with pytest.raises(nx.NetworkXAlgorithmError):
            bipartite.birank(G, all_nodes)

    @pytest.mark.parametrize(
        "damping_factor,value", itertools.product(["alpha", "beta"], [-0.1, 1.1])
    )
    def test_birank_invalid_alpha_beta(self, damping_factor, value):
        kwargs = {damping_factor: value}
        with pytest.raises(nx.NetworkXAlgorithmError):
            bipartite.birank(
                self.davis_southern_women_graph, self.women_bipartite_set, **kwargs
            )

    def test_birank_power_iteration_failed_convergence(self):
        with pytest.raises(nx.PowerIterationFailedConvergence):
            bipartite.birank(
                self.davis_southern_women_graph, self.women_bipartite_set, max_iter=1
            )

    @pytest.mark.parametrize(
        "personalization,alpha,beta",
        itertools.product(
            [
                # Concentrated case
                lambda x: 1000 if x == 0 else 0,
                # Uniform case
                lambda x: 5,
                # Zero case
                lambda x: 0,
            ],
            [i / 2 for i in range(3)],
            [i / 2 for i in range(3)],
        ),
    )
    def test_gnmk_convergence_birank(self, personalization, alpha, beta):
        top_personalization_dict = {
            node: personalization(node) for node in self.gnmk_random_graph_top_nodes
        }
        bipartite.birank(
            self.gnmk_random_graph,
            self.gnmk_random_graph_top_nodes,
            top_personalization=top_personalization_dict,
            alpha=alpha,
            beta=beta,
        )

    def test_negative_personalization(self):
        top_personalization_dict = {0: -1}
        with pytest.raises(nx.NetworkXAlgorithmError):
            bipartite.birank(
                self.gnmk_random_graph,
                self.gnmk_random_graph_top_nodes,
                top_personalization=top_personalization_dict,
            )



"""
================
Visibility Graph
================

Visibility Graph constructed from a time series
"""

import matplotlib.pyplot as plt

import networkx as nx

time_series = [0, 2, 1, 3, 2, 1, 3, 2, 1, 3, 2, 1, 3, 4, 0]
# or
# import random
# time_series = [random.randint(1, 10) for i in range(10)]

G = nx.visibility_graph(time_series)

labels = nx.get_node_attributes(G, "value")

fig, all_axes = plt.subplots(2, 1, num="Visibility Graph", figsize=(8, 12))
axs = all_axes.flat

layouts_params = {
    # a layout emphasizing the line-of-sight connectivity
    "Line-of-Sight Connectivity": {
        "pos": {x: (x, 0) for x in range(len(time_series))},
        "connectionstyle": "arc3,rad=-1.57079632679",
    },
    # a layout showcasing the time series values
    "Time Series values with Connectivity": {
        "pos": {i: (i, v) for i, v in enumerate(time_series)}
    },
}

for i, (name, params) in enumerate(layouts_params.items()):
    axs[i].title.set_text(name)
    axs[i].title.set_size(11)
    axs[i].set_xlabel("Time", size=10)
    axs[i].margins(0.10)
    nx.draw_networkx_nodes(G, params.get("pos"), ax=axs[i], alpha=0.5)
    nx.draw_networkx_labels(G, params.get("pos"), ax=axs[i], labels=labels)
    nx.draw_networkx_edges(
        G, **params, ax=axs[i], arrows=True, arrowstyle="<->", arrowsize=10
    )

axs[1].set_ylabel("Value", size=10)

fig.suptitle("Visibility Graph")
fig.tight_layout()
plt.show()



"""
===============
Basic Animation
===============

NetworkX supports several 3D layout functions for visualization.
These layouts can be combined with :external+matplotlib:doc:`api/animation_api`
for 3D graph visualization.

This example shows a basic animation incrementing the camera view.
"""

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from matplotlib import animation

###############################################################################
# Define a graph to plot.
# -----------------------
#
# Pick a graph for visualization in 3D, such as one of the Platonic solids

G = nx.dodecahedral_graph()
pos = nx.spectral_layout(G, dim=3)
nodes = np.array([pos[v] for v in G])
edges = np.array([(pos[u], pos[v]) for u, v in G.edges()])

###############################################################################
# Rotating 3D graph animation.
# ----------------------------
#
# In this example, a frame update is only a rotation of a given 3D graph.


def init():
    ax.clear()
    ax.scatter(*nodes.T, alpha=0.2, s=100, color="blue")
    for vizedge in edges:
        ax.plot(*vizedge.T, color="gray")
    ax.grid(False)
    ax.set_axis_off()


def _frame_update(index):
    ax.view_init(index * 0.2, index * 0.5)


fig = plt.figure()
ax = fig.add_subplot(111, projection="3d")
fig.tight_layout()

ani = animation.FuncAnimation(
    fig,
    _frame_update,
    init_func=init,
    interval=50,
    cache_frame_data=False,
    frames=100,
)

plt.show()



# See https://github.com/networkx/networkx/pull/1474
# Copyright 2011 Reya Group <http://www.reyagroup.com>
# Copyright 2011 Alex Levenson <alex@isnotinvain.com>
# Copyright 2011 Diederik van Liere <diederik.vanliere@rotman.utoronto.ca>
"""Functions for analyzing triads of a graph."""

from collections import defaultdict
from itertools import combinations, permutations

import networkx as nx
from networkx.utils import not_implemented_for, py_random_state

__all__ = [
    "triadic_census",
    "is_triad",
    "all_triads",
    "triads_by_type",
    "triad_type",
]

#: The integer codes representing each type of triad.
#:
#: Triads that are the same up to symmetry have the same code.
TRICODES = (
    1,
    2,
    2,
    3,
    2,
    4,
    6,
    8,
    2,
    6,
    5,
    7,
    3,
    8,
    7,
    11,
    2,
    6,
    4,
    8,
    5,
    9,
    9,
    13,
    6,
    10,
    9,
    14,
    7,
    14,
    12,
    15,
    2,
    5,
    6,
    7,
    6,
    9,
    10,
    14,
    4,
    9,
    9,
    12,
    8,
    13,
    14,
    15,
    3,
    7,
    8,
    11,
    7,
    12,
    14,
    15,
    8,
    14,
    13,
    15,
    11,
    15,
    15,
    16,
)

#: The names of each type of triad. The order of the elements is
#: important: it corresponds to the tricodes given in :data:`TRICODES`.
TRIAD_NAMES = (
    "003",
    "012",
    "102",
    "021D",
    "021U",
    "021C",
    "111D",
    "111U",
    "030T",
    "030C",
    "201",
    "120D",
    "120U",
    "120C",
    "210",
    "300",
)


#: A dictionary mapping triad code to triad name.
TRICODE_TO_NAME = {i: TRIAD_NAMES[code - 1] for i, code in enumerate(TRICODES)}


def _tricode(G, v, u, w):
    """Returns the integer code of the given triad.

    This is some fancy magic that comes from Batagelj and Mrvar's paper. It
    treats each edge joining a pair of `v`, `u`, and `w` as a bit in
    the binary representation of an integer.

    """
    combos = ((v, u, 1), (u, v, 2), (v, w, 4), (w, v, 8), (u, w, 16), (w, u, 32))
    return sum(x for u, v, x in combos if v in G[u])


@not_implemented_for("undirected")
@nx._dispatchable
def triadic_census(G, nodelist=None):
    """Determines the triadic census of a directed graph.

    The triadic census is a count of how many of the 16 possible types of
    triads are present in a directed graph. If a list of nodes is passed, then
    only those triads are taken into account which have elements of nodelist in them.

    Parameters
    ----------
    G : digraph
       A NetworkX DiGraph
    nodelist : list
        List of nodes for which you want to calculate triadic census

    Returns
    -------
    census : dict
       Dictionary with triad type as keys and number of occurrences as values.

    Examples
    --------
    >>> G = nx.DiGraph([(1, 2), (2, 3), (3, 1), (3, 4), (4, 1), (4, 2)])
    >>> triadic_census = nx.triadic_census(G)
    >>> for key, value in triadic_census.items():
    ...     print(f"{key}: {value}")
    003: 0
    012: 0
    102: 0
    021D: 0
    021U: 0
    021C: 0
    111D: 0
    111U: 0
    030T: 2
    030C: 2
    201: 0
    120D: 0
    120U: 0
    120C: 0
    210: 0
    300: 0

    Notes
    -----
    This algorithm has complexity $O(m)$ where $m$ is the number of edges in
    the graph.

    For undirected graphs, the triadic census can be computed by first converting
    the graph into a directed graph using the ``G.to_directed()`` method.
    After this conversion, only the triad types 003, 102, 201 and 300 will be
    present in the undirected scenario.

    Raises
    ------
    ValueError
        If `nodelist` contains duplicate nodes or nodes not in `G`.
        If you want to ignore this you can preprocess with `set(nodelist) & G.nodes`

    See also
    --------
    triad_graph

    References
    ----------
    .. [1] Vladimir Batagelj and Andrej Mrvar, A subquadratic triad census
        algorithm for large sparse networks with small maximum degree,
        University of Ljubljana,
        http://vlado.fmf.uni-lj.si/pub/networks/doc/triads/triads.pdf

    """
    nodeset = set(G.nbunch_iter(nodelist))
    if nodelist is not None and len(nodelist) != len(nodeset):
        raise ValueError("nodelist includes duplicate nodes or nodes not in G")

    N = len(G)
    Nnot = N - len(nodeset)  # can signal special counting for subset of nodes

    # create an ordering of nodes with nodeset nodes first
    m = {n: i for i, n in enumerate(nodeset)}
    if Nnot:
        # add non-nodeset nodes later in the ordering
        not_nodeset = G.nodes - nodeset
        m.update((n, i + N) for i, n in enumerate(not_nodeset))

    # build all_neighbor dicts for easy counting
    # After Python 3.8 can leave off these keys(). Speedup also using G._pred
    # nbrs = {n: G._pred[n].keys() | G._succ[n].keys() for n in G}
    nbrs = {n: G.pred[n].keys() | G.succ[n].keys() for n in G}
    dbl_nbrs = {n: G.pred[n].keys() & G.succ[n].keys() for n in G}

    if Nnot:
        sgl_nbrs = {n: G.pred[n].keys() ^ G.succ[n].keys() for n in not_nodeset}
        # find number of edges not incident to nodes in nodeset
        sgl = sum(1 for n in not_nodeset for nbr in sgl_nbrs[n] if nbr not in nodeset)
        sgl_edges_outside = sgl // 2
        dbl = sum(1 for n in not_nodeset for nbr in dbl_nbrs[n] if nbr not in nodeset)
        dbl_edges_outside = dbl // 2

    # Initialize the count for each triad to be zero.
    census = {name: 0 for name in TRIAD_NAMES}
    # Main loop over nodes
    for v in nodeset:
        vnbrs = nbrs[v]
        dbl_vnbrs = dbl_nbrs[v]
        if Nnot:
            # set up counts of edges attached to v.
            sgl_unbrs_bdy = sgl_unbrs_out = dbl_unbrs_bdy = dbl_unbrs_out = 0
        for u in vnbrs:
            if m[u] <= m[v]:
                continue
            unbrs = nbrs[u]
            neighbors = (vnbrs | unbrs) - {u, v}
            # Count connected triads.
            for w in neighbors:
                if m[u] < m[w] or (m[v] < m[w] < m[u] and v not in nbrs[w]):
                    code = _tricode(G, v, u, w)
                    census[TRICODE_TO_NAME[code]] += 1

            # Use a formula for dyadic triads with edge incident to v
            if u in dbl_vnbrs:
                census["102"] += N - len(neighbors) - 2
            else:
                census["012"] += N - len(neighbors) - 2

            # Count edges attached to v. Subtract later to get triads with v isolated
            # _out are (u,unbr) for unbrs outside boundary of nodeset
            # _bdy are (u,unbr) for unbrs on boundary of nodeset (get double counted)
            if Nnot and u not in nodeset:
                sgl_unbrs = sgl_nbrs[u]
                sgl_unbrs_bdy += len(sgl_unbrs & vnbrs - nodeset)
                sgl_unbrs_out += len(sgl_unbrs - vnbrs - nodeset)
                dbl_unbrs = dbl_nbrs[u]
                dbl_unbrs_bdy += len(dbl_unbrs & vnbrs - nodeset)
                dbl_unbrs_out += len(dbl_unbrs - vnbrs - nodeset)
        # if nodeset == G.nodes, skip this b/c we will find the edge later.
        if Nnot:
            # Count edges outside nodeset not connected with v (v isolated triads)
            census["012"] += sgl_edges_outside - (sgl_unbrs_out + sgl_unbrs_bdy // 2)
            census["102"] += dbl_edges_outside - (dbl_unbrs_out + dbl_unbrs_bdy // 2)

    # calculate null triads: "003"
    # null triads = total number of possible triads - all found triads
    total_triangles = (N * (N - 1) * (N - 2)) // 6
    triangles_without_nodeset = (Nnot * (Nnot - 1) * (Nnot - 2)) // 6
    total_census = total_triangles - triangles_without_nodeset
    census["003"] = total_census - sum(census.values())

    return census


@nx._dispatchable
def is_triad(G):
    """Returns True if the graph G is a triad, else False.

    Parameters
    ----------
    G : graph
       A NetworkX Graph

    Returns
    -------
    istriad : boolean
       Whether G is a valid triad

    Examples
    --------
    >>> G = nx.DiGraph([(1, 2), (2, 3), (3, 1)])
    >>> nx.is_triad(G)
    True
    >>> G.add_edge(0, 1)
    >>> nx.is_triad(G)
    False
    """
    if isinstance(G, nx.Graph):
        if G.order() == 3 and nx.is_directed(G):
            if not any((n, n) in G.edges() for n in G.nodes()):
                return True
    return False


@not_implemented_for("undirected")
@nx._dispatchable(returns_graph=True)
def all_triads(G):
    """A generator of all possible triads in G.

    Parameters
    ----------
    G : digraph
       A NetworkX DiGraph

    Returns
    -------
    all_triads : generator of DiGraphs
       Generator of triads (order-3 DiGraphs)

    Examples
    --------
    >>> G = nx.DiGraph([(1, 2), (2, 3), (3, 1), (3, 4), (4, 1), (4, 2)])
    >>> for triad in nx.all_triads(G):
    ...     print(triad.edges)
    [(1, 2), (2, 3), (3, 1)]
    [(1, 2), (4, 1), (4, 2)]
    [(3, 1), (3, 4), (4, 1)]
    [(2, 3), (3, 4), (4, 2)]

    """
    triplets = combinations(G.nodes(), 3)
    for triplet in triplets:
        yield G.subgraph(triplet).copy()


@not_implemented_for("undirected")
@nx._dispatchable
def triads_by_type(G):
    """Returns a list of all triads for each triad type in a directed graph.
    There are exactly 16 different types of triads possible. Suppose 1, 2, 3 are three
    nodes, they will be classified as a particular triad type if their connections
    are as follows:

    - 003: 1, 2, 3
    - 012: 1 -> 2, 3
    - 102: 1 <-> 2, 3
    - 021D: 1 <- 2 -> 3
    - 021U: 1 -> 2 <- 3
    - 021C: 1 -> 2 -> 3
    - 111D: 1 <-> 2 <- 3
    - 111U: 1 <-> 2 -> 3
    - 030T: 1 -> 2 -> 3, 1 -> 3
    - 030C: 1 <- 2 <- 3, 1 -> 3
    - 201: 1 <-> 2 <-> 3
    - 120D: 1 <- 2 -> 3, 1 <-> 3
    - 120U: 1 -> 2 <- 3, 1 <-> 3
    - 120C: 1 -> 2 -> 3, 1 <-> 3
    - 210: 1 -> 2 <-> 3, 1 <-> 3
    - 300: 1 <-> 2 <-> 3, 1 <-> 3

    Refer to the :doc:`example gallery </auto_examples/graph/plot_triad_types>`
    for visual examples of the triad types.

    Parameters
    ----------
    G : digraph
       A NetworkX DiGraph

    Returns
    -------
    tri_by_type : dict
       Dictionary with triad types as keys and lists of triads as values.

    Examples
    --------
    >>> G = nx.DiGraph([(1, 2), (1, 3), (2, 3), (3, 1), (5, 6), (5, 4), (6, 7)])
    >>> dict = nx.triads_by_type(G)
    >>> dict["120C"][0].edges()
    OutEdgeView([(1, 2), (1, 3), (2, 3), (3, 1)])
    >>> dict["012"][0].edges()
    OutEdgeView([(1, 2)])

    References
    ----------
    .. [1] Snijders, T. (2012). "Transitivity and triads." University of
        Oxford.
        https://web.archive.org/web/20170830032057/http://www.stats.ox.ac.uk/~snijders/Trans_Triads_ha.pdf
    """
    # num_triads = o * (o - 1) * (o - 2) // 6
    # if num_triads > TRIAD_LIMIT: print(WARNING)
    all_tri = all_triads(G)
    tri_by_type = defaultdict(list)
    for triad in all_tri:
        name = triad_type(triad)
        tri_by_type[name].append(triad)
    return tri_by_type


@not_implemented_for("undirected")
@nx._dispatchable
def triad_type(G):
    """Returns the sociological triad type for a triad.

    Parameters
    ----------
    G : digraph
       A NetworkX DiGraph with 3 nodes

    Returns
    -------
    triad_type : str
       A string identifying the triad type

    Examples
    --------
    >>> G = nx.DiGraph([(1, 2), (2, 3), (3, 1)])
    >>> nx.triad_type(G)
    '030C'
    >>> G.add_edge(1, 3)
    >>> nx.triad_type(G)
    '120C'

    Notes
    -----
    There can be 6 unique edges in a triad (order-3 DiGraph) (so 2^^6=64 unique
    triads given 3 nodes). These 64 triads each display exactly 1 of 16
    topologies of triads (topologies can be permuted). These topologies are
    identified by the following notation:

    {m}{a}{n}{type} (for example: 111D, 210, 102)

    Here:

    {m}     = number of mutual ties (takes 0, 1, 2, 3); a mutual tie is (0,1)
              AND (1,0)
    {a}     = number of asymmetric ties (takes 0, 1, 2, 3); an asymmetric tie
              is (0,1) BUT NOT (1,0) or vice versa
    {n}     = number of null ties (takes 0, 1, 2, 3); a null tie is NEITHER
              (0,1) NOR (1,0)
    {type}  = a letter (takes U, D, C, T) corresponding to up, down, cyclical
              and transitive. This is only used for topologies that can have
              more than one form (eg: 021D and 021U).

    References
    ----------
    .. [1] Snijders, T. (2012). "Transitivity and triads." University of
        Oxford.
        https://web.archive.org/web/20170830032057/http://www.stats.ox.ac.uk/~snijders/Trans_Triads_ha.pdf
    """
    if not is_triad(G):
        raise nx.NetworkXAlgorithmError("G is not a triad (order-3 DiGraph)")
    num_edges = len(G.edges())
    if num_edges == 0:
        return "003"
    elif num_edges == 1:
        return "012"
    elif num_edges == 2:
        e1, e2 = G.edges()
        if set(e1) == set(e2):
            return "102"
        elif e1[0] == e2[0]:
            return "021D"
        elif e1[1] == e2[1]:
            return "021U"
        elif e1[1] == e2[0] or e2[1] == e1[0]:
            return "021C"
    elif num_edges == 3:
        for e1, e2, e3 in permutations(G.edges(), 3):
            if set(e1) == set(e2):
                if e3[0] in e1:
                    return "111U"
                # e3[1] in e1:
                return "111D"
            elif set(e1).symmetric_difference(set(e2)) == set(e3):
                if {e1[0], e2[0], e3[0]} == {e1[0], e2[0], e3[0]} == set(G.nodes()):
                    return "030C"
                # e3 == (e1[0], e2[1]) and e2 == (e1[1], e3[1]):
                return "030T"
    elif num_edges == 4:
        for e1, e2, e3, e4 in permutations(G.edges(), 4):
            if set(e1) == set(e2):
                # identify pair of symmetric edges (which necessarily exists)
                if set(e3) == set(e4):
                    return "201"
                if {e3[0]} == {e4[0]} == set(e3).intersection(set(e4)):
                    return "120D"
                if {e3[1]} == {e4[1]} == set(e3).intersection(set(e4)):
                    return "120U"
                if e3[1] == e4[0]:
                    return "120C"
    elif num_edges == 5:
        return "210"
    elif num_edges == 6:
        return "300"
